"""
Equations for the High Velocity Impact Problems
###############################################
"""

from math import sqrt
from pysph.sph.equation import Equation

class VonMisesPlasticity2D(Equation):
    def __init__(self, dest, sources, flow_stress):
        self.flow_stress2 = float(flow_stress*flow_stress)
        self.factor = sqrt( 2.0/3.0 )*flow_stress
        super(VonMisesPlasticity2D,self).__init__(dest, sources)

    def loop(self, d_idx, d_s00, d_s01, d_s02, d_s11, d_s12, d_s22):
        s00a = d_s00[d_idx]
        s01a = d_s01[d_idx]
        s02a = d_s02[d_idx]

        s10a = d_s01[d_idx]
        s11a = d_s11[d_idx]
        s12a = d_s12[d_idx]

        s20a = d_s02[d_idx]
        s21a = d_s12[d_idx]
        s22a = d_s22[d_idx]

        J = s00a* s00a + 2.0 * s01a*s10a + s11a*s11a
        scale = 1.0
        if (J > 2.0/3.0 * self.flow_stress2):
            scale = self.factor/sqrt(J)

        # store the stresses
        d_s00[d_idx] = scale * s00a
        d_s01[d_idx] = scale * s01a
        d_s02[d_idx] = scale * s02a

        d_s11[d_idx] = scale * s11a
        d_s12[d_idx] = scale * s12a

        d_s22[d_idx] = scale * s22a


class StiffenedGasEOS(Equation):

    """Stiffened-gas EOS from "A Free Lagrange Augmented Godunov Method for the
    Simulation of Elastic-Plastic Solids", B. P. Howell and G. J. Ball, JCP
    (2002).  http://dx.doi.org/10.1006/jcph.2001.6931
    """

    def __init__(self, dest, sources, gamma, r0, c0):
        self.gamma = float(gamma) # Gruneisen gamma
        self.c0    = float(c0)    # unshocked sound speed
        self.r0    = float(r0)    # reference density

        super(StiffenedGasEOS,self).__init__(dest, sources)

    def loop(self, d_idx, d_e, d_rho, d_p, d_cs):
        # Eq. (17)
        d_p[d_idx] = self.c0*self.c0 * (d_rho[d_idx] - self.r0) + \
                      (self.gamma - 1.0)*d_rho[d_idx]*d_e[d_idx]

        # Eq. (21)
        d_cs[d_idx] = sqrt(
            self.c0*self.c0 + (self.gamma - 1.0) *
            (d_e[d_idx] + d_p[d_idx]/d_rho[d_idx])
        )


class MieGruneisenEOS(Equation):
    def __init__(self, dest, sources, gamma,r0, c0, S):

        self.gamma = float(gamma)
        self.r0 = float(r0)
        self.c0 = float(c0)
        self.S = float(S)

        self.a0 = a0 = float(r0 * c0 * c0)
        self.b0 = a0 * ( 1 + 2.0*(S - 1.0) )
        self.c0 = a0 * ( 2*(S - 1.0) + 3*(S - 1.0)*(S - 1.0) )

        super(MieGruneisenEOS, self).__init__(dest, sources)

    def loop(self, d_idx, d_p, d_rho, d_e):
        rhoa = d_rho[d_idx]
        ea = d_e[d_idx]

        gamma = self.gamma
        ratio = rhoa/self.r0 - 1.0
        ratio2 = ratio * ratio

        PH = self.a0 * ratio
        if ratio > 0:
            PH = PH + ratio2 * (self.b0 + self.c0*ratio)

        d_p[d_idx] = (1. - 0.5*gamma*ratio) * PH + rhoa * ea * gamma



# cython: language_level=3, embedsignature=True
# distutils: language=c++

# Numpy
cimport numpy as np

from cpython cimport dict
from cpython cimport list

# PyZoltan
from cyarray.carray cimport UIntArray, IntArray, DoubleArray, LongArray
from pyzoltan.core.zoltan cimport PyZoltan, ZoltanGeometricPartitioner
from pyzoltan.core.zoltan_comm cimport ZComm
from pyzoltan.czoltan.czoltan_types cimport ZOLTAN_ID_TYPE, ZOLTAN_ID_PTR, ZOLTAN_OK

# PySPH imports
from pysph.base.nnps_base cimport NNPSParticleArrayWrapper
from pysph.base.particle_array cimport ParticleArray
from pysph.base.point cimport *

cdef class ParticleArrayExchange:
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef public int msglength_tag_remote           # msg length tag for remote_exchange
    cdef public int data_tag_remote                # data tag for remote_exchange

    cdef public int msglength_tag_lb               # msg length tag for lb_exchange
    cdef public int data_tag_lb                    # data tag for lb_exchange

    cdef public int pa_index                       # Particle index
    cdef public ParticleArray pa                   # Particle data
    cdef public NNPSParticleArrayWrapper pa_wrapper    # wrapper to exchange data

    # flags to indicate whether data needs to be exchanged
    cdef public bint lb_exchange
    cdef public bint remote_exchange

    cdef public size_t num_local         # Total number of particles
    cdef public size_t num_global        # Global number of particles
    cdef public size_t num_remote        # Number of remote particles
    cdef public size_t num_ghost         # Number of ghost particles


    # mpi.Comm object and associated rank and size
    cdef public object comm
    cdef public int rank, size

    # list of load balancing props
    cdef public list lb_props
    cdef public int nprops
    cdef public double lb_weight

    # Import/Export lists for particles
    cdef public UIntArray exportParticleGlobalids
    cdef public UIntArray exportParticleLocalids
    cdef public IntArray exportParticleProcs
    cdef public int numParticleExport

    cdef public UIntArray importParticleGlobalids
    cdef public UIntArray importParticleLocalids
    cdef public IntArray importParticleProcs
    cdef public int numParticleImport

    ############################################################################
    # Member functions
    ############################################################################
    # exchange data given send and receive lists
    cdef exchange_data(self, ZComm zcomm, dict sendbufs, int count)

# base class for all parallel managers
cdef class ParallelManager:
    ############################################################################
    # Data Attributes
    ############################################################################
    # mpi comm, rank and size
    cdef public object comm
    cdef public int rank
    cdef public int size

    cdef public np.ndarray dt_sendbuf    # array for global reduction
    cdef public np.ndarray dt_recvbuf    # for time step calculation

    cdef public int lb_freq              # load balancing frequency
    cdef public int lb_count             # counter for current lb step

    cdef public int ncells_local         # number of local cells
    cdef public int ncells_remote        # number of remote cells
    cdef public int ncells_total         # total number of cells
    cdef public list cell_list           # list of cells

    cdef public dict cell_map            # index structure
    cdef public int ghost_layers         # BOunding box size
    cdef public double cell_size         # cell size used for binning

    # list of particle arrays, wrappers, exchange and nnps instances
    cdef public list particles
    cdef public list pa_wrappers
    cdef public list pa_exchanges

    # number of local and remote particles
    cdef public list num_local
    cdef public list num_remote
    cdef public list num_global

    cdef public double radius_scale      # Radius scale for kernel

    cdef public bint initial_update
    cdef public bint update_cell_sizes

    # number of arrays
    cdef int narrays

    # boolean for parallel
    cdef bint in_parallel

    # Min and max across all processors
    cdef np.ndarray minx, miny, minz     # min and max arrays used
    cdef np.ndarray maxx, maxy, maxz     # for MPI Allreduce
    cdef np.ndarray maxh                 # operations

    cdef public double mx, my, mz        # global min and max values
    cdef public double Mx, My, Mz, Mh

    # global indices for the cells
    cdef UIntArray cell_gid

    # cell coordinate values
    cdef DoubleArray cx, cy, cz

    # Import/Export lists for cells
    cdef public UIntArray exportCellGlobalids
    cdef public UIntArray exportCellLocalids
    cdef public IntArray exportCellProcs
    cdef public int numCellExport

    cdef public UIntArray importCellGlobalids
    cdef public UIntArray importCellLocalids
    cdef public IntArray importCellProcs
    cdef public int numCellImport

    ############################################################################
    # Member functions
    ############################################################################
    # Index particles given by a list of indices. The indices are
    # assumed to be of type unsigned int and local to the NNPS object
    cdef _bin(self, int pa_index, UIntArray indices)

    # Compute the cell size across processors. The cell size is taken
    # as max(h)*radius_scale
    cpdef compute_cell_size(self)

    # compute global bounds for the particle distribution. The bounds
    # are the min and max coordinate values across all processors and
    # the maximum smoothing length needed for parallel binning.
    cdef _compute_bounds(self)

    # nearest neighbor search routines taking into account multiple
    # particle arrays
    cpdef get_nearest_particles(self, int src_index, int dst_index,
                                size_t d_idx, UIntArray nbrs)

# Zoltan based parallel cell manager for SPH simulations
cdef class ZoltanParallelManager(ParallelManager):
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef public int changes              # logical (0,1) if the partition changes
    cdef public PyZoltan pz              # the PyZoltan wrapper for lb etc

# Class of geometric load balancers
cdef class ZoltanParallelManagerGeometric(ZoltanParallelManager):
    pass



import os
import tempfile
import shutil
import subprocess
import sys

from pytest import mark

from pysph.examples import run


def check_output(*args, **kw):
    """Simple hack to support Python 2.6 which does not have
    subprocess.check_output.
    """
    if not hasattr(subprocess, 'check_output'):
        subprocess.call(*args, **kw)
    else:
        subprocess.check_output(*args, **kw)


def print_safe(string_or_bytes):
    if type(string_or_bytes) is bytes:
        print(string_or_bytes.decode('utf-8'))
    else:
        print(string_or_bytes)


_orig_ets_toolkit = None
def setup_module():
    # Set the ETS_TOOLKIT to null to avoid errors when importing TVTK.
    global _orig_ets_toolkit
    var = 'ETS_TOOLKIT'
    _orig_ets_toolkit = os.environ.get(var)
    os.environ[var] = 'null'


def teardown_module():
    var = 'ETS_TOOLKIT'
    if _orig_ets_toolkit is None:
        del os.environ[var]
    else:
        os.environ[var] = _orig_ets_toolkit


def run_example(module):
    """This simply runs the example to make sure that the example executes
    correctly.  It wipes out the generated output directory.
    """
    out_dir = tempfile.mkdtemp()
    cmd = [sys.executable, "-m", module, "--max-steps", "1",
           "--disable-output", "-q", "-d", out_dir]
    env_vars = dict(os.environ)
    env_vars['ETS_TOOLKIT'] = 'null'
    try:
        check_output(cmd, env=env_vars)
    except subprocess.CalledProcessError as e:
        print_safe(e.stdout)
        print_safe(e.stderr)
        raise
    finally:
        shutil.rmtree(out_dir)


def _has_tvtk():
    try:
        from tvtk.api import tvtk
    except (ImportError, SystemExit):
        return False
    else:
        return True


def _find_examples():
    examples = []
    for module, doc in run.get_all_examples():
        if module == 'pysph.examples.rigid_body.dam_break3D_sph' and \
           not _has_tvtk():
                continue
        examples.append(module)
    return examples


@mark.slow
@mark.parametrize("module", _find_examples())
def test_example_should_run(module):
    run_example(module)



# cython: language_level=3, embedsignature=True
# distutils: language=c++

# Library imports.
import numpy as np
cimport numpy as np

# Cython imports
from cython.operator cimport dereference as deref, preincrement as inc
from cython.parallel import parallel, prange, threadid

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.vector cimport vector

cdef extern from "<algorithm>" namespace "std" nogil:
    void sort[Iter, Compare](Iter first, Iter last, Compare comp)
    void sort[Iter](Iter first, Iter last)

# cpython
from cpython.dict cimport PyDict_Clear, PyDict_Contains, PyDict_GetItem
from cpython.list cimport PyList_GetItem, PyList_SetItem, PyList_GET_ITEM

# Cython for compiler directives
cimport cython

import pyopencl as cl
import pyopencl.array
from pyopencl.elementwise import ElementwiseKernel

from pysph.base.nnps_base cimport *
from pysph.base.device_helper import DeviceHelper
from compyle.config import get_config
from compyle.array import get_backend, Array
from compyle.parallel import Elementwise, Scan
from compyle.types import annotate
from compyle.opencl import (get_context, get_queue,
                            set_context, set_queue)
import compyle.array as array

# Particle Tag information
from cyarray.carray cimport BaseArray, aligned_malloc, aligned_free
from .utils import ParticleTAGS

from .nnps_base cimport *

from pysph.base.gpu_helper_kernels import (exclusive_input, exclusive_output,
                                           get_scan)


cdef class GPUNeighborCache:
    def __init__(self, GPUNNPS nnps, int dst_index, int src_index,
            backend=None):
        self.backend = get_backend(backend)
        self._dst_index = dst_index
        self._src_index = src_index
        self._nnps = nnps
        self._particles = nnps.particles
        self._narrays = nnps.narrays
        cdef long n_p = self._particles[dst_index].get_number_of_particles()

        self._get_start_indices = None
        self._cached = False
        self._copied_to_cpu = False

        self._nbr_lengths_gpu = Array(np.uint32, n=n_p,
                                      backend=self.backend)

        self._neighbors_gpu = Array(np.uint32,
                                    backend=self.backend)

    #### Public protocol ################################################

    cdef void get_neighbors_raw_gpu(self):
        if not self._cached:
            self._find_neighbors()

    cdef void get_neighbors_raw(self, size_t d_idx, UIntArray nbrs):
        self.get_neighbors_raw_gpu()
        if not self._copied_to_cpu:
            self.copy_to_cpu()
        nbrs.c_reset()
        nbrs.c_set_view(self._neighbors_cpu_ptr + self._start_idx_ptr[d_idx],
                self._nbr_lengths_ptr[d_idx])

    #### Private protocol ################################################

    cdef void _find_neighbors(self):

        self._nnps.find_neighbor_lengths(self._nbr_lengths_gpu)
        # FIXME:
        # - Store sum kernel
        # - don't allocate neighbors_gpu each time.
        # - Don't allocate _nbr_lengths and start_idx.

        total_size_gpu = array.sum(self._nbr_lengths_gpu)

        cdef unsigned long total_size = <unsigned long>(total_size_gpu)

        # Allocate _neighbors_cpu and neighbors_gpu
        self._neighbors_gpu.resize(total_size)

        self._start_idx_gpu = self._nbr_lengths_gpu.copy()

        # Do prefix sum on self._neighbor_lengths for the self._start_idx
        if self._get_start_indices is None:
            self._get_start_indices = get_scan(exclusive_input, exclusive_output,
                    dtype=np.uint32, backend=self.backend)

        self._get_start_indices(ary=self._start_idx_gpu)

        self._nnps.find_nearest_neighbors_gpu(self._neighbors_gpu,
                self._start_idx_gpu)
        self._cached = True

    cdef void copy_to_cpu(self):
        self._copied_to_cpu = True
        self._neighbors_cpu = self._neighbors_gpu.get()
        self._neighbors_cpu_ptr = <unsigned int*> self._neighbors_cpu.data
        self._nbr_lengths = self._nbr_lengths_gpu.get()
        self._nbr_lengths_ptr = <unsigned int*> self._nbr_lengths.data
        self._start_idx = self._start_idx_gpu.get()
        self._start_idx_ptr = <unsigned int*> self._start_idx.data

    cpdef update(self):
        # FIXME: Don't allocate here unless needed.
        self._cached = False
        self._copied_to_cpu = False
        cdef long n_p = self._particles[self._dst_index].get_number_of_particles()
        self._nbr_lengths_gpu.resize(n_p)

    cpdef get_neighbors(self, int src_index, size_t d_idx, UIntArray nbrs):
        self.get_neighbors_raw(d_idx, nbrs)

    cpdef get_neighbors_gpu(self):
        self.get_neighbors_raw_gpu()


cdef class GPUNNPS(NNPSBase):
    """Nearest neighbor query class using the box-sort algorithm.

    NNPS bins all local particles using the box sort algorithm in
    Cells. The cells are stored in a dictionary 'cells' which is keyed
    on the spatial index (IntPoint) of the cell.

    """
    def __init__(self, int dim, list particles, double radius_scale=2.0,
                 int ghost_layers=1, domain=None, bint cache=True,
                 bint sort_gids=False, backend=None):
        """Constructor for NNPS

        Parameters
        ----------

        dim : int
            Dimension (fixme: Not sure if this is really needed)

        particles : list
            The list of particle arrays we are working on.

        radius_scale : double, default (2)
            Optional kernel radius scale. Defaults to 2

        domain : DomainManager, default (None)
            Optional limits for the domain

        cache : bint
            Flag to set if we want to cache neighbor calls. This costs
            storage but speeds up neighbor calculations.

        sort_gids : bint, default (False)
            Flag to sort neighbors using gids (if they are available).
            This is useful when comparing parallel results with those
            from a serial run.

        backend : string
            Backend on which to build NNPS Module
        """
        NNPSBase.__init__(self, dim, particles, radius_scale, ghost_layers,
                domain, cache, sort_gids)

        self.backend = get_backend(backend)
        self.backend = 'opencl' if self.backend is 'cython' else self.backend
        self.use_double = get_config().use_double
        self.dtype = np.float64 if self.use_double else np.float32
        self.dtype_max = np.finfo(self.dtype).max
        self._last_domain_size = 0.0

        # Set the device helper if needed.
        for pa in particles:
            if pa.gpu is None:
                pa.set_device_helper(DeviceHelper(pa, backend=self.backend))

        # The cache.
        self.use_cache = cache
        _cache = []
        for d_idx in range(len(particles)):
            for s_idx in range(len(particles)):
                _cache.append(GPUNeighborCache(self, d_idx, s_idx,
                    backend=self.backend))
        self.cache = _cache
        self.use_double = get_config().use_double

    cdef void get_nearest_neighbors(self, size_t d_idx, UIntArray nbrs):
        if self.use_cache:
            self.current_cache.get_neighbors_raw(d_idx, nbrs)
        else:
            nbrs.c_reset()
            self.find_nearest_neighbors(d_idx, nbrs)

    cpdef get_nearest_particles_gpu(self, int src_index, int dst_index):
        cdef int idx = dst_index*self.narrays + src_index
        if self.src_index != src_index \
            or self.dst_index != dst_index:
            self.set_context(src_index, dst_index)
        self.cache[idx].get_neighbors_gpu()

    cpdef spatially_order_particles(self, int pa_index):
        """Spatially order particles such that nearby particles have indices
        nearer each other.  This may improve pre-fetching on the CPU.
        """
        indices, callback = self.get_spatially_ordered_indices(pa_index)
        self.particles[pa_index].gpu.align(indices)
        callback()

    def set_use_cache(self, bint use_cache):
        self.use_cache = use_cache
        if use_cache:
            for cache in self.cache:
                cache.update()

    cdef void find_neighbor_lengths(self, nbr_lengths):
        raise NotImplementedError("NNPS :: find_neighbor_lengths called")

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices):
        raise NotImplementedError("NNPS :: find_nearest_neighbors called")

    cpdef update(self):
        cdef int i, num_particles
        cdef ParticleArray pa

        cdef DomainManager domain = self.domain

        # use cell sizes computed by the domain.
        self.cell_size = domain.manager.cell_size
        self.hmin = domain.manager.hmin

        # compute bounds and refresh the data structure
        self._compute_bounds()
        self._refresh()

        # indices on which to bin. We bin all local particles
        for i in range(self.narrays):
            # bin the particles
            self._bin(pa_index=i)

        if self.use_cache:
            for cache in self.cache:
                cache.update()

    def update_domain(self):
        self.domain.update()

    cdef _compute_bounds(self):
        """Compute coordinate bounds for the particles"""
        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef double domain_size
        xmax = -self.dtype_max
        ymax = -self.dtype_max
        zmax = -self.dtype_max

        xmin = self.dtype_max
        ymin = self.dtype_max
        zmin = self.dtype_max

        for pa_wrapper in pa_wrappers:
            x = pa_wrapper.pa.gpu.get_device_array('x')
            y = pa_wrapper.pa.gpu.get_device_array('y')
            z = pa_wrapper.pa.gpu.get_device_array('z')

            pa_wrapper.pa.gpu.update_minmax_cl(['x', 'y', 'z'])

            # find min and max of variables
            xmax = np.maximum(x.maximum, xmax)
            ymax = np.maximum(y.maximum, ymax)
            zmax = np.maximum(z.maximum, zmax)

            xmin = np.minimum(x.minimum, xmin)
            ymin = np.minimum(y.minimum, ymin)
            zmin = np.minimum(z.minimum, zmin)

        # Add a small offset to the limits.
        lx, ly, lz = xmax - xmin, ymax - ymin, zmax - zmin
        xmin -= lx*0.01; ymin -= ly*0.01; zmin -= lz*0.01
        xmax += lx*0.01; ymax += ly*0.01; zmax += lz*0.01

        domain_size = fmax(lx, ly)
        domain_size = fmax(domain_size, lz)
        if self._last_domain_size > 1e-16 and \
           domain_size > 2.0*self._last_domain_size:
            msg = (
                '*'*70 +
                '\nWARNING: Domain size has increased by a large amount.\n' +
                'Particles are probably diverging, please check your code!\n' +
                '*'*70
            )
            print(msg)
            self._last_domain_size = domain_size

        # If all of the dimensions have very small extent give it a unit size.
        _eps = 1e-12
        if (np.abs(xmax - xmin) < _eps) and (np.abs(ymax - ymin) < _eps) \
            and (np.abs(zmax - zmin) < _eps):
            xmin -= 0.5; xmax += 0.5
            ymin -= 0.5; ymax += 0.5
            zmin -= 0.5; zmax += 0.5

        # store the minimum and maximum of physical coordinates
        self.xmin = np.asarray([xmin, ymin, zmin])
        self.xmax = np.asarray([xmax, ymax, zmax])

    cpdef _bin(self, int pa_index):
        raise NotImplementedError("NNPS :: _bin called")

    cpdef _refresh(self):
        raise NotImplementedError("NNPS :: _refresh called")


cdef class BruteForceNNPS(GPUNNPS):
    def __init__(self, int dim, list particles, double radius_scale=2.0,
            int ghost_layers=1, domain=None, bint cache=True,
            bint sort_gids=False, backend='opencl'):
        GPUNNPS.__init__(self, dim, particles, radius_scale, ghost_layers,
                domain, cache, sort_gids, backend)

        self.radius_scale2 = radius_scale*radius_scale
        self.src_index = -1
        self.dst_index = -1
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

        cdef str norm2 = \
                """
                #define NORM2(X, Y, Z) ((X)*(X) + (Y)*(Y) + (Z)*(Z))
                """

        self.preamble = norm2

    cpdef set_context(self, int src_index, int dst_index):
        """Setup the context before asking for neighbors.  The `dst_index`
        represents the particles for whom the neighbors are to be determined
        from the particle array with index `src_index`.

        Parameters
        ----------

         src_index: int: the source index of the particle array.
         dst_index: int: the destination index of the particle array.
        """
        GPUNNPS.set_context(self, src_index, dst_index)

        self.src = self.pa_wrappers[ src_index ]
        self.dst = self.pa_wrappers[ dst_index ]

    cdef void find_neighbor_lengths(self, nbr_lengths):
        # IMPORTANT NOTE: pyopencl uses the length of the first argument
        # to determine the global work size
        arguments = \
                """%(data_t)s* d_x, %(data_t)s* d_y, %(data_t)s* d_z,
                %(data_t)s* d_h, %(data_t)s* s_x, %(data_t)s* s_y,
                %(data_t)s* s_z, %(data_t)s* s_h, unsigned int num_particles,
                unsigned int* nbr_lengths, %(data_t)s radius_scale2
                """ % {"data_t" : ("double" if self.use_double else "float")}

        src = """
                unsigned int j;
                unsigned int length = 0;
                %(data_t)s dist;
                %(data_t)s h_i = radius_scale2*d_h[i]*d_h[i];
                %(data_t)s h_j;
                for(j=0; j<num_particles; j++)
                {
                    h_j = radius_scale2*s_h[j]*s_h[j];
                    dist = NORM2(d_x[i] - s_x[j], d_y[i] - s_y[j], d_z[i] - s_z[j]);
                    if(dist < h_i || dist < h_j)
                        length += 1;
                }
                nbr_lengths[i] = length;
                """ % {"data_t" : ("double" if self.use_double else "float")}

        brute_force_nbr_lengths = ElementwiseKernel(
            get_context(), arguments, src, "brute_force_nbr_lengths",
            preamble=self.preamble
        )

        src_gpu = self.src.pa.gpu
        dst_gpu = self.dst.pa.gpu
        brute_force_nbr_lengths(dst_gpu.x.dev, dst_gpu.y.dev, dst_gpu.z.dev,
                dst_gpu.h.dev, src_gpu.x.dev, src_gpu.y.dev, src_gpu.z.dev,
                src_gpu.h.dev, self.src.get_number_of_particles(),
                nbr_lengths.dev, self.radius_scale2)

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices):

        arguments = \
                """%(data_t)s* d_x, %(data_t)s* d_y, %(data_t)s* d_z,
                %(data_t)s* d_h, %(data_t)s* s_x, %(data_t)s* s_y,
                %(data_t)s* s_z, %(data_t)s* s_h, unsigned int num_particles,
                unsigned int* start_indices, unsigned int* nbrs,
                %(data_t)s radius_scale2
                """ % {"data_t" : ("double" if self.use_double else "float")}

        src = """
                unsigned int j, k = 0;
                %(data_t)s dist;
                %(data_t)s h_i = radius_scale2*d_h[i]*d_h[i];
                %(data_t)s h_j;
                unsigned int idx = start_indices[i];
                for(j=0; j<num_particles; j++)
                {
                    h_j = radius_scale2*s_h[j]*s_h[j];
                    dist = NORM2(d_x[i] - s_x[j], d_y[i] - s_y[j], d_z[i] - s_z[j]);
                    if(dist < h_i || dist < h_j)
                    {
                        nbrs[idx + k] = j;
                        k += 1;
                    }
                }
                """ % {"data_t" : ("double" if self.use_double else "float")}

        brute_force_nbrs = ElementwiseKernel(
            get_context(), arguments, src, "brute_force_nbrs",
            preamble=self.preamble
        )

        src_gpu = self.src.pa.gpu
        dst_gpu = self.dst.pa.gpu

        brute_force_nbrs(dst_gpu.x.dev, dst_gpu.y.dev, dst_gpu.z.dev,
                dst_gpu.h.dev, src_gpu.x.dev, src_gpu.y.dev, src_gpu.z.dev,
                src_gpu.h.dev, self.src.get_number_of_particles(),
                start_indices.dev, nbrs.dev, self.radius_scale2)

    cpdef _bin(self, int pa_index):
        pass

    cpdef _refresh(self):
        pass



# cython: language_level=3, embedsignature=True
# distutils: language=c++

from .nnps_base cimport *
from .octree cimport Octree, CompressedOctree, cOctreeNode

from libcpp.vector cimport vector
cimport cython

ctypedef unsigned int u_int

cdef extern from 'math.h':
    int abs(int) nogil
    double ceil(double) nogil
    double floor(double) nogil
    double fabs(double) nogil
    double fmax(double, double) nogil
    double fmin(double, double) nogil

cdef class OctreeNNPS(NNPS):
    ##########################################################################
    # Data Attributes
    ##########################################################################
    cdef list tree
    cdef cOctreeNode* current_tree
    cdef u_int* current_pids

    cdef double radius_scale2
    cdef NNPSParticleArrayWrapper dst, src
    cdef int leaf_max_particles

    ##########################################################################
    # Member functions
    ##########################################################################

    cpdef get_depth(self, int pa_index)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil

    cpdef set_context(self, int src_index, int dst_index)

    cdef void _get_neighbors(self, double q_x, double q_y, double q_z, double q_h,
            double* src_x_ptr, double* src_y_ptr, double* src_z_ptr, double* src_h_ptr,
            UIntArray nbrs, cOctreeNode* node) nogil

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices)

    cpdef _refresh(self)

    cpdef _bin(self, int pa_index, UIntArray indices)

cdef class CompressedOctreeNNPS(OctreeNNPS):
    ##########################################################################
    # Member functions
    ##########################################################################
    cpdef set_context(self, int src_index, int dst_index)

    cpdef _refresh(self)

    cpdef _bin(self, int pa_index, UIntArray indices)



"""Manage the PySPH and Compyle cache directories.

These directories contain the generated sources and extension modules and can
get quite big. The command allows you to see the path and size of these cache
directories and also clear them out if they are too big.

"""
import argparse
from pathlib import Path
import shutil
import sys


def _get_cache_dirs():
    home = Path('~').expanduser()
    cc = home / '.compyle' / 'source'
    pc = home / '.pysph' / 'source'
    return (cc, pc)


def _find_size(pth):
    return sum(f.stat().st_size for f in pth.glob('**/*') if f.is_file())


def show_cache():
    cc, pc = _get_cache_dirs()
    print("PySPH cache directories are at:")
    GB = 2**30
    print("{}  {:<.3g} GB".format(str(cc), _find_size(cc)/GB))
    print("{}    {:<.3g} GB".format(str(pc), _find_size(pc)/GB))


def clear_cache():
    cc, pc = _get_cache_dirs()
    print("Clearing cache in\n", cc, "\n", pc)
    confirm = input('Are you sure? (y/N) ')
    if confirm in ['y', 'Y']:
        shutil.rmtree(cc)
        shutil.rmtree(pc)


def main(argv=None):
    parser = argparse.ArgumentParser(
        prog='cache', description=__doc__, add_help=False
    )

    parser.add_argument(
        "-h",
        "--help",
        action="store_true",
        default=False,
        dest="help",
        help="show this help message and exit"
    )

    parser.add_argument(
        "-c", "--clear",
        action="store_true",
        default=False,
        help="Delete all the files in the cache directory."
    )

    if argv is not None and len(argv) > 0 and argv[0] in ['-h', '--help']:
        parser.print_help()
        sys.exit()

    options, extra = parser.parse_known_args(argv)
    if options.clear:
        clear_cache()
    else:
        show_cache()


if __name__ == '__main__':
    main()



"""Culls files in a given directory, one in every 'c' files is spared.

The specified directory can contain other directories that house the output
 files; files in all these directories will be culled, sparing one in every
 'c' files. Note that DELETION IS PERMANENT.
"""

from pysph.tools.binder import find_sim_dirs, find_dir_size
from pysph.solver.utils import get_files
import os
import sys
import argparse


def cull(src_path, c):

    src_path = os.path.abspath(src_path)
    sim_paths_list = find_sim_dirs(src_path)

    initial_size = find_dir_size(src_path)

    for path in sim_paths_list:
        files = get_files(path)
        l = len(files)
        del_files = [
            files[i] for i in set(range(l)) - set(range(0, l, c))
        ]
        if len(del_files) != 0:
            for f in del_files:
                os.remove(f)

    final_size = find_dir_size(src_path)

    print("Initial size of the directory was: "+str(initial_size)+" bytes")
    print("Final size of the directory is: "+str(final_size)+" bytes")
    return


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='cull',
        description=__doc__,
        add_help=False
    )

    parser.add_argument(
        "-h",
        "--help",
        action="store_true",
        default=False,
        dest="help",
        help="show this help message and exit"
    )

    parser.add_argument(
        "src_path",
        metavar='src_path',
        type=str,
        nargs=1,
        help="the directory containing the directories/files to be culled"
    )

    parser.add_argument(
        "-c", "--cull-factor",
        metavar="cull_factor",
        type=int,
        default=2,
        help="one in every 'c' files is spared, all remaining output " +
             "files are deleted [default=2]"
    )

    if len(argv) > 0 and argv[0] in ['-h', '--help']:
        parser.print_help()
        sys.exit()

    options, extra = parser.parse_known_args(argv)

    src_path = options.src_path[0]
    x = options.cull_factor

    cull(src_path, x)


if __name__ == '__main__':
    main()



#cython: embedsignature=True
import numpy as np
cimport numpy as np
from pysph.base.tree.point_tree import PointTree

cdef class OctreeGPUNNPS(GPUNNPS):
    def __init__(self, int dim, list particles, double radius_scale=2.0,
                 int ghost_layers=1, domain=None, bint fixed_h=False,
                 bint cache=True, bint sort_gids=False,
                 allow_sort=False, leaf_size=32,
                 bint use_elementwise=False, bint use_partitions=False,
                 backend=None):
        GPUNNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids, backend=backend
        )

        self.src_index = -1
        self.dst_index = -1
        self.sort_gids = sort_gids
        self.leaf_size = leaf_size

        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int i, num_particles

        self.octrees = []

        if self.use_double:
            from warnings import warn
            warn("Octree NNPS by default uses single precision arithmetic for"
                 "finding neighbors. A few particles outside of the original "
                 "search radius might be included.")

        for i in range(self.narrays):
            self.octrees.append(PointTree(pa=self.pa_wrappers[i].pa,
                                          radius_scale=radius_scale,
                                          use_double=self.use_double,
                                          leaf_size=leaf_size, dim=dim))
        self.use_elementwise = use_elementwise
        self.use_partitions = use_partitions
        self.allow_sort = allow_sort
        self.domain.update()
        self.update()

        # Check if device supports required workgroup size,
        # else default to elementwise nnps
        if not self.octrees[0]._is_valid_nnps_wgs():
            from warnings import warn
            warn("Octree NNPS with given leaf size (%d) is "
                 "not supported for given device. Switching to a elementwise "
                 "version of the Octree NNPS" % leaf_size)

            self.use_elementwise = True

    cpdef _bin(self, int pa_index):
        self.octrees[pa_index].refresh(self.xmin, self.xmax,
                                       self.domain.manager.hmin)
        self.octrees[pa_index].set_node_bounds()

        if self.allow_sort:
            self.spatially_order_particles(pa_index)

    def get_spatially_ordered_indices(self, int pa_index):
        def update():
            self.octrees[pa_index]._sort()

        return self.octrees[pa_index].pids, update

    cpdef _refresh(self):
        pass

    cpdef set_context(self, int src_index, int dst_index):
        """Setup the context before asking for neighbors.  The `dst_index`
        represents the particles for whom the neighbors are to be determined
        from the particle array with index `src_index`.

        Parameters
        ----------

         src_index: int: the source index of the particle array.
         dst_index: int: the destination index of the particle array.
        """
        GPUNNPS.set_context(self, src_index, dst_index)

        self.src_index = src_index
        self.dst_index = dst_index

        octree_src = self.octrees[src_index]
        octree_dst = self.octrees[dst_index]
        self.dst_src = src_index != dst_index

        self.neighbor_cid_counts, self.neighbor_cids = octree_dst.find_neighbor_cids(
            octree_src)

    cdef void find_neighbor_lengths(self, nbr_lengths):
        octree_src = self.octrees[self.src_index]
        octree_dst = self.octrees[self.dst_index]

        args = []
        if self.use_elementwise:
            find_neighbor_lengths = octree_dst.find_neighbor_lengths_elementwise
        else:
            find_neighbor_lengths = octree_dst.find_neighbor_lengths

        find_neighbor_lengths(
            self.neighbor_cid_counts, self.neighbor_cids, octree_src,
            nbr_lengths, *args
        )

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices):
        octree_src = self.octrees[self.src_index]
        octree_dst = self.octrees[self.dst_index]

        args = []
        if self.use_elementwise:
            find_neighbors = octree_dst.find_neighbors_elementwise
        else:
            find_neighbors = octree_dst.find_neighbors
            args.append(self.use_partitions)

        find_neighbors(
            self.neighbor_cid_counts, self.neighbor_cids, octree_src,
            start_indices, nbrs, *args
        )

    cpdef get_kernel_args(self, c_type):
        octree_dst = self.octrees[self.dst_index]
        octree_src = self.octrees[self.src_index]
        pa_gpu_dst = octree_dst.pa.gpu
        pa_gpu_src = octree_src.pa.gpu

        return [
                   octree_dst.unique_cids.dev.data,
                   octree_src.pids.dev.data,
                   octree_dst.pids.dev.data,
                   octree_dst.cids.dev.data,
                   octree_src.pbounds.dev.data, octree_dst.pbounds.dev.data,
                   np.float32(octree_dst.radius_scale),
                   self.neighbor_cid_counts.dev.data,
                   self.neighbor_cids.dev.data,
                   pa_gpu_dst.x.dev.data, pa_gpu_dst.y.dev.data,
                   pa_gpu_dst.z.dev.data, pa_gpu_dst.h.dev.data,
                   pa_gpu_src.x.dev.data, pa_gpu_src.y.dev.data,
                   pa_gpu_src.z.dev.data, pa_gpu_src.h.dev.data
               ], [
                   (self.leaf_size * octree_dst.unique_cid_count,),
                   (self.leaf_size,)
               ]



import numpy as np
from stl import stl

# PySPH base and carray imports
from pysph.base.utils import get_particle_array
from pysph.base.kernels import QuinticSpline
from pysph.solver.solver import Solver
from pysph.sph.integrator import EulerIntegrator, IntegratorStep
from pysph.sph.integrator_step import EulerStep
from pysph.sph.equation import Group, Equation
from pysph.solver.application import Application
from pysph.tools.geometry import remove_overlap_particles

from pysph.tools.particle_packing import (
    ParticlePacking, calculate_normal_2d_surface, shift_surface_inside)


def get_packing_folders(folder, dx):
    """Get all the required folder and files names for the packing.
    This is done to avoid mixing of multiple particle spacing

     Parameters
    ----------
    folder : name of the output folder from Application subclass
    dx : required particles spacing

    Returns
    -------
    preprocess_folder : The output folder for packing Application
    layer_folder : The output folder for layer packing around the hexagonal
                   packing
    res_file : the name of the packed file containing the coordinates
    """
    import os
    parent = os.path.dirname(folder)
    basename = os.path.basename(folder)
    preprocess_folder = os.path.join(parent, 'packing_%.4f' % dx)
    layer_folder = os.path.join(parent, 'layer_%.4f' % dx)

    res_folder = os.path.join(parent, 'preprocess')
    os.makedirs(res_folder, exist_ok=True)

    res_file = os.path.join(parent, 'preprocess', basename + '_%.4f.npz' % dx)

    return preprocess_folder, layer_folder, res_file


def readdata(resfile):
    data = np.load(resfile)
    xs = data['xs']
    ys = data['ys']
    zs = data['zs']
    xf = data['xf']
    yf = data['yf']
    zf = data['zf']
    return xs, ys, zs, xf, yf, zf


class Packer(Application):
    def __init__(self, fname, output_dir, domain, add_opt_func, dx, out,
                 dim=None, x=None, y=None, z=None, L=0.0, B=0.0,
                 H=0.0, filename=None,
                 hardpoints=None, use_prediction=False,
                 filter_layers=False, reduce_dfreq=False,
                 tol=1e-2, scale=1.0, shift=False,
                 invert_normal=False, pb=None, nu=None,
                 k=None, dfreq=-1, no_solid=False):
        self.hdx = 1.2
        self.dx = dx
        self.x = x
        self.y = y
        self.z = z
        self.L = L
        self.B = B
        self.H = H
        self.filename = filename
        self.dfreq = dfreq
        self.hardpoints = {} if hardpoints is None else hardpoints
        self.use_prediction = use_prediction
        self.filter_layers = filter_layers
        self.reduce_dfreq = reduce_dfreq
        self.tol = tol
        self.dim = dim
        self.scale = scale
        self.shift = shift
        self.invert_normal = invert_normal
        self.pb = pb
        self.nu = nu
        self.k = k
        self.out = out
        self.no_solid = no_solid
        self.add_opt_func = add_opt_func

        self.bound = self._get_bound()

        super(Packer, self).__init__(fname, output_dir, domain)

    def add_user_options(self, group):
        self.add_opt_func(group)

    def _get_bound(self):
        import sys
        import os
        from pysph.tools.particle_packing import get_bounding_box
        if self.filename is not None:
            file, ext = os.path.splitext(self.filename)
            print(ext)
            if (ext == '.txt') or (ext == '.csv'):
                self.dim = 2
            elif ext == '.stl':
                self.dim = 3
            else:
                print('file extension %s not supported' % ext)
                sys.exit()

            if self.dim == 2:
                try:
                    self.x, self.y = np.loadtxt(self.filename, unpack=True)
                    self.x *= self.scale
                    self.y *= self.scale
                except IOError:
                    print('read the file')
                    print('The supported file format is \"x y\"')
                    sys.exit()

            if self.dim == 3:
                try:
                    data = stl.StlMesh(self.filename)
                    self.x = self.scale * data.x
                    self.y = self.scale * data.y
                    self.z = self.scale * data.z
                except IOError:
                    print('read the file')
                    print('The supported file format is \"x y\"')
                    sys.exit()

        if self.z is None:
            self.dim = 2
            self.z = np.zeros_like(self.x)

        return get_bounding_box(
            self.dx, self.x, self.y, self.z, self.L, self.B, self.H)

    def create_particles(self):
        s = self.scheme

        bound = self.bound
        free = s.create_free_particles(bound, name='free')
        frozen = s.create_frozen_container(bound, name='frozen')
        particles = [free, frozen]
        if (self.filename is None) and (self.x is None):
            free = s.create_free_particles(bound, name='free', rect=True)
            frozen = s.create_frozen_container(bound, name='frozen', rect=True)
            np.random.seed(10)
            x = free.x
            free.x += (np.random.random(len(x)) - 0.5) * self.dx
            free.y += (np.random.random(len(x)) - 0.5) * self.dx
            if self.dim == 3:
                free.z += (np.random.random(len(x)) - 0.5) * self.dx
            particles = [free, frozen]
        else:
            nodes = None
            if self.filename is None:
                nodes = s.create_boundary_node(
                    self.filename, [self.x, self.y], scale=self.scale,
                    shift=self.shift, invert=self.invert_normal,
                    name='nodes')
            else:
                nodes = s.create_boundary_node(
                    self.filename, scale=self.scale, shift=self.shift,
                    invert=self.invert_normal, name='nodes')

            boundary = get_particle_array(name='boundary')
            particles.extend([boundary, nodes])

        s.setup_properties(particles)
        for pa in particles:
            pa.dt_adapt[:] = 1e20
        return particles

    def create_scheme(self):
        hardpoints = self.hardpoints
        if self.no_solid:
            s = ParticlePacking(
                fluids=['free'], solids={}, frozen=['frozen'],
                dim=self.dim, hdx=self.hdx, dx=self.dx, nu=self.nu,
                pb=self.pb, k=self.k, tol=self.tol)
        else:
            s = ParticlePacking(
                fluids=['free'], solids={'boundary': 'nodes'},
                frozen=['frozen'], dim=self.dim,
                use_prediction=self.use_prediction,
                filter_layers=self.filter_layers,
                reduce_dfreq=self.reduce_dfreq,
                hdx=self.hdx, dx=self.dx, hardpoints=hardpoints,
                nu=self.nu, pb=self.pb, k=self.k, tol=self.tol,
                dfreq=self.dfreq)

        s.configure_solver(dt=1e-5)
        return s

    def post_step(self, solver):
        self.scheme.post_step(self.particles, solver)

    def post_process(self, info_fname):
        import os
        from pysph.solver.utils import load
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return
        res = self.out
        filename = self.output_files[-1]
        data = load(filename)
        free = data['arrays']['free']
        frozen = data['arrays']['frozen']
        solid = None
        solid_nodes = None
        if len(data['arrays']) > 2:
            solid = data['arrays']['boundary']
            solid_nodes = data['arrays']['nodes']
        self.scheme.post_process(
            free, solid, solid_nodes, frozen, self.dx, res)


class HexaToRectLayer(Packer):
    def create_particles(self):
        s = self.scheme

        bound = self.bound
        free = s.create_free_particles(bound, name='free', outer=True)
        frozen = s.create_frozen_container(bound, name='frozen', outer=True)
        particles = [free, frozen]
        s.setup_properties(particles)
        for pa in particles:
            pa.dt_adapt[:] = 1e20
        return particles



"""
SPH Boundary Equations
######################
"""

from pysph.sph.equation import Equation

def wendland_quintic(rij=1.0, h=1.0):
    q = rij/h
    q1 = 2.0 - q
    val = 0.0
    if q < 2.0:
        val = (1 + 2.5*q + 2*q*q)*q1*q1*q1*q1*q1

    return val


class MonaghanBoundaryForce(Equation):
    def __init__(self, dest, sources, deltap):
        self.deltap = deltap
        super(MonaghanBoundaryForce,self).__init__(dest,sources)

    def loop(self, d_idx, s_idx, s_m, s_rho, d_m, d_cs, s_cs, d_h,
             s_tx, s_ty, s_tz, s_nx, s_ny, s_nz,
             d_au, d_av, d_aw, XIJ):

        norm = declare('matrix((3,))')
        tang = declare('matrix((3,))')

        ma = d_m[d_idx]
        mb = s_m[s_idx]

        # particle sound speed
        cs = d_cs[d_idx]

        # boundary normals
        norm[0] = s_nx[s_idx]
        norm[1] = s_ny[s_idx]
        norm[2] = s_nz[s_idx]

        # boundary tangents
        tang[0] = s_tx[s_idx]
        tang[1] = s_ty[s_idx]
        tang[2] = s_tz[s_idx]

        # x and y projections
        x = XIJ[0]*tang[0] + XIJ[1]*tang[1] + XIJ[2]*tang[2]
        y = XIJ[0]*norm[0] + XIJ[1]*norm[1] + XIJ[2]*norm[2]

        # compute the force
        force = 0.0
        q = y/d_h[d_idx]

        xabs = fabs(x)

        if (0 <= xabs) and (xabs <= self.deltap):
            beta = 0.02 * cs * cs/y
            tforce = 1.0 - xabs/self.deltap

            if (0 < q) and (q <= 2.0/3.0):
                nforce =  2.0/3.0

            elif (2.0/3.0 < q) and (q <= 1.0):
                nforce = 2*q*(1.0 - 0.75*q)

            elif (1.0 < q) and (q <= 2.0):
                nforce = 0.5 * (2-q)*(2-q)

            else:
                nforce = 0.0

            force = (mb/(ma+mb)) * nforce * tforce * beta
        else:
            force = 0.0

        # boundary force accelerations
        d_au[d_idx] += force * norm[0]
        d_av[d_idx] += force * norm[1]
        d_aw[d_idx] += force * norm[2]

class MonaghanKajtarBoundaryForce(Equation):
    def __init__(self, dest, sources, K=None, beta=None, h=None):
        self.K = K
        self.beta = beta
        self.h = h

        if None in [K, beta, h]:
            raise ValueError("Invalid parameter values")

        super(MonaghanKajtarBoundaryForce,self).__init__(dest,sources)

    def _get_helpers_(self):
        return [wendland_quintic]

    def loop(self, d_idx, s_idx, d_m, s_m, d_au, d_av, d_aw, RIJ, R2IJ, XIJ):

        ma = d_m[d_idx]
        mb = s_m[s_idx]

        w = wendland_quintic(RIJ, self.h)
        force = self.K/self.beta * w/R2IJ * 2*mb/(ma + mb)

        d_au[d_idx] += force * XIJ[0]
        d_av[d_idx] += force * XIJ[1]
        d_aw[d_idx] += force * XIJ[2]



"""Convenience script for running various PySPH related tasks.
"""

from __future__ import print_function

from argparse import ArgumentParser
from os.path import exists, join
import sys


def run_viewer(args):
    from pysph.tools.mayavi_viewer import main
    main(args)


def run_examples(args):
    from pysph.examples.run import main
    main(args)


def output_vtk(args):
    from pysph.solver.vtk_output import main
    main(args)


def _has_pysph_dir():
    init_py = join('pysph', '__init__.py')
    init_pyc = join('pysph', '__init__.pyc')
    return exists(init_py) or exists(init_pyc)


def run_tests(args):
    argv = ['--pyargs', 'pysph'] + args
    from pytest import cmdline
    cmdline.main(args=argv)


def make_binder(args):
    from pysph.tools.binder import main
    main(args)


def cull_files(args):
    from pysph.tools.cull import main
    main(args)


def manage_cache(args):
    from pysph.tools.manage_cache import main
    main(args)


def main():
    parser = ArgumentParser(description=__doc__, add_help=False)
    parser.add_argument(
        "-h", "--help", action="store_true", default=False, dest="help",
        help="show this help message and exit"
    )
    subparsers = parser.add_subparsers(help='sub-command help')

    viewer = subparsers.add_parser(
        'view', help='View output files generated by PySPH',
        add_help=False
    )
    viewer.set_defaults(func=run_viewer)

    runner = subparsers.add_parser(
        'run', help='Run PySPH examples',
        add_help=False
    )
    runner.set_defaults(func=run_examples)

    vtk_out = subparsers.add_parser(
        'dump_vtk', help='Dump VTK Output',
        add_help=False
    )
    vtk_out.set_defaults(func=output_vtk)

    tests = subparsers.add_parser(
        'test', help='Run entire PySPH test-suite',
        add_help=False
    )
    tests.set_defaults(func=run_tests)

    binder = subparsers.add_parser(
        'binder',
        help='Make a mybinder.org compatible directory for upload to a ' +
             'host repo',
        add_help=False
    )
    binder.set_defaults(func=make_binder)

    cull = subparsers.add_parser(
        'cull',
        help='Cull files in a given directory by a specified culling_factor',
        add_help=False
    )
    cull.set_defaults(func=cull_files)

    cache = subparsers.add_parser(
        'cache',
        help='Show cache directories or clear them',
        add_help=False
    )
    cache.set_defaults(func=manage_cache)

    if (len(sys.argv) == 1 or (len(sys.argv) > 1 and
                               sys.argv[1] in ['-h', '--help'])):
        parser.print_help()
        sys.exit()

    args, extra = parser.parse_known_args()
    args.func(extra)


if __name__ == '__main__':
    main()



try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import numpy
from .particle_array import ParticleArray, \
    get_local_tag, get_remote_tag, get_ghost_tag

from cyarray.api import LongArray

UINT_MAX = (1 << 32) - 1


# Internal tags used in PySPH (defined in particle_array.pxd)
class ParticleTAGS:
    Local = get_local_tag()
    Remote = get_remote_tag()
    Ghost = get_ghost_tag()


def arange_long(start, stop=-1):
    """ Creates a LongArray working same as builtin range with upto 2 arguments
    both expected to be positive
    """

    if stop == -1:
        arange = LongArray(start)
        for i in range(start):
            arange.data[i] = i
        return arange
    else:
        size = stop-start
        arange = LongArray(size)
        for i in range(size):
            arange.data[i] = start + i
        return arange


# A collection of default properties for all SPH arrays.
DEFAULT_PROPS = set(
    ('x', 'y', 'z', 'u', 'v', 'w', 'm', 'h', 'rho', 'p',
     'au', 'av', 'aw', 'gid', 'pid', 'tag')
)


def get_particle_array(additional_props=None, constants=None, backend=None,
                       **props):
    """Create and return a particle array with default properties.

    The default properties are ['x', 'y', 'z', 'u', 'v', 'w', 'm', 'h', 'rho',
    'p', 'au', 'av', 'aw', 'gid', 'pid', 'tag'], this set is available in
    `DEFAULT_PROPS`.


    Parameters
    ----------

    additional_props : list
        If specified, add these properties.

    constants : dict
        Any constants to be added to the particle array.

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    Examples
    --------

    >>> x = linspace(0,1,10)
    >>> pa = get_particle_array(name='fluid', x=x)
    >>> pa.properties.keys()
    ['x', 'z', 'rho', 'pid', 'v', 'tag', 'm', 'p', 'gid', 'au',
     'aw', 'av', 'y', 'u', 'w', 'h']
    >>> pa1 = get_particle_array(name='fluid', additional_props=['xx', 'yy'])

    >>> pa = get_particle_array(name='fluid', x=x, constants={'alpha': 1.0})
    >>> pa.constants.keys()
    ['alpha']

    """

    # handle the name separately
    if 'name' in props:
        name = props['name']
        props.pop('name')
    else:
        name = "array"

    # default properties for an SPH particle
    default_props = set(DEFAULT_PROPS)

    # add any additional props to the default_props
    if additional_props:
        default_props = default_props.union(additional_props)

    np = 0

    prop_dict = {}
    for prop in props.keys():
        data = numpy.asarray(props[prop])
        np = data.size

        if prop in ['tag', 'pid']:
            prop_dict[prop] = {'data': data,
                               'type': 'int',
                               'name': prop}
        elif prop in ['gid']:
            prop_dict[prop] = {'data': data.astype(numpy.uint32),
                               'type': 'unsigned int',
                               'name': prop}
        else:
            prop_dict[prop] = {'data': data,
                               'type': 'double',
                               'name': prop}

    # Add the default props
    for prop in default_props:
        if prop not in prop_dict:
            if prop in ["pid"]:
                prop_dict[prop] = {'name': prop, 'type': 'int',
                                   'default': 0}
            elif prop in ['tag']:
                prop_dict[prop] = {'name': prop, 'type': 'int',
                                   'default': ParticleTAGS.Local}
            elif prop in ['gid']:
                data = numpy.ones(shape=np, dtype=numpy.uint32)
                data[:] = UINT_MAX

                prop_dict[prop] = {'name': prop, 'type': 'unsigned int',
                                   'data': data, 'default': UINT_MAX}

            else:
                prop_dict[prop] = {'name': prop, 'type': 'double',
                                   'default': 0}

    # create the particle array
    pa = ParticleArray(name=name, constants=constants, backend=backend,
                       **prop_dict)

    # default property arrays to save out. Any reasonable SPH particle
    # should define these
    pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'm', 'h',
                          'pid', 'gid', 'tag'])

    return pa


def get_particle_array_wcsph(constants=None, **props):
    """Return a particle array for the WCSPH formulation.

    This sets the default properties to be::

        ['x', 'y', 'z', 'u', 'v', 'w', 'h', 'rho', 'm', 'p', 'cs', 'ax', 'ay',
        'az', 'au', 'av', 'aw', 'x0','y0', 'z0','u0', 'v0','w0', 'arho',
        'rho0', 'div', 'gid','pid', 'tag']

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """

    wcsph_props = ['cs', 'ax', 'ay', 'az', 'arho', 'x0', 'y0', 'z0',
                   'u0', 'v0', 'w0', 'rho0', 'div', 'dt_cfl', 'dt_force']

    pa = get_particle_array(
        constants=constants, additional_props=wcsph_props, **props
    )

    # default property arrays to save out.
    pa.set_output_arrays([
        'x', 'y', 'z', 'u', 'v', 'w', 'rho', 'm', 'h',
        'pid', 'gid', 'tag', 'p'
    ])

    return pa


def get_particle_array_iisph(constants=None, **props):
    """Get a particle array for the IISPH formulation.

    The default properties are::

        ['x', 'y', 'z', 'u', 'v', 'w', 'm', 'h', 'rho', 'p', 'au', 'av', 'aw',
        'gid', 'pid', 'tag' 'uadv', 'vadv', 'wadv', 'rho_adv', 'au', 'av',
        'aw','ax', 'ay', 'az', 'dii0', 'dii1', 'dii2', 'V', 'aii', 'dijpj0',
        'dijpj1', 'dijpj2', 'p', 'p0', 'piter', 'compression'
         ]

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """
    iisph_props = ['uadv', 'vadv', 'wadv', 'rho_adv',
                   'au', 'av', 'aw', 'ax', 'ay', 'az',
                   'dii0', 'dii1', 'dii2', 'V', 'dt_cfl', 'dt_force',
                   'aii', 'dijpj0', 'dijpj1', 'dijpj2', 'p', 'p0', 'piter',
                   'compression']
    # Used to calculate the total compression first index is count and second
    # the compression.
    consts = {'tmp_comp': [0.0, 0.0]}
    if constants:
        consts.update(constants)

    pa = get_particle_array(
        constants=consts, additional_props=iisph_props, **props
    )
    pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'h', 'm',
                          'p', 'pid', 'au', 'av', 'aw', 'tag', 'gid', 'V'])
    return pa


def get_particle_array_rigid_body(constants=None, **props):
    """Return a particle array for a rigid body motion.

    For multiple bodies, add a body_id property starting at index 0 with each
    index denoting the body to which the particle corresponds to.

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """
    extra_props = ['au', 'av', 'aw', 'V', 'fx', 'fy', 'fz', 'x0', 'y0', 'z0',
                   'tang_disp_x', 'tang_disp_y', 'tang_disp_z', 'tang_disp_x0',
                   'tang_disp_y0', 'tang_disp_z0', 'tang_velocity_x',
                   'tang_velocity_y', 'rad_s',
                   'tang_velocity_z',  'nx', 'ny', 'nz']

    body_id = props.pop('body_id', None)
    nb = 1 if body_id is None else numpy.max(body_id) + 1

    consts = {'total_mass': numpy.zeros(nb, dtype=float),
              'num_body': numpy.asarray(nb, dtype=int),
              'cm': numpy.zeros(3*nb, dtype=float),

              # The mi are also used to temporarily reduce mass (1), center of
              # mass (3) and the interia components (6), total force (3), total
              # torque (3).
              'mi': numpy.zeros(16*nb, dtype=float),
              'force': numpy.zeros(3*nb, dtype=float),
              'torque': numpy.zeros(3*nb, dtype=float),
              # velocity, acceleration of CM.
              'vc': numpy.zeros(3*nb, dtype=float),
              'ac': numpy.zeros(3*nb, dtype=float),
              'vc0': numpy.zeros(3*nb, dtype=float),
              # angular velocity, acceleration of body.
              'omega': numpy.zeros(3*nb, dtype=float),
              'omega0': numpy.zeros(3*nb, dtype=float),
              'omega_dot': numpy.zeros(3*nb, dtype=float)
              }
    if constants:
        consts.update(constants)
    pa = get_particle_array(constants=consts, additional_props=extra_props,
                            **props)
    pa.add_property('body_id', type='int', data=body_id)
    pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'h', 'm',
                          'p', 'pid', 'au', 'av', 'aw', 'tag', 'gid', 'V',
                          'fx', 'fy', 'fz', 'body_id'])
    return pa


def get_particle_array_tvf_fluid(constants=None, **props):
    """Return a particle array for the TVF formulation for a fluid.

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """
    tv_props = ['uhat', 'vhat', 'what',
                'auhat', 'avhat', 'awhat', 'vmag2', 'V']

    pa = get_particle_array(
        constants=constants, additional_props=tv_props, **props
    )
    pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'p', 'h',
                          'm', 'au', 'av', 'aw', 'V', 'vmag2', 'pid', 'gid',
                          'tag'])

    return pa


def get_particle_array_tvf_solid(constants=None, **props):
    """Return a particle array for the TVF formulation for a solid.

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """
    tv_props = ['u0', 'v0', 'w0', 'V', 'wij', 'ax', 'ay', 'az',
                'uf', 'vf', 'wf', 'ug', 'vg', 'wg']

    pa = get_particle_array(
        constants=constants, additional_props=tv_props, **props
    )
    pa.set_output_arrays(
        ['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'p', 'h', 'm', 'V',
         'pid', 'gid', 'tag']
    )
    return pa


def get_particle_array_gasd(constants=None, **props):
    """Return a particle array for a Gas Dynamics problem.

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """
    required_props = [
        'x', 'y', 'z', 'u', 'v', 'w', 'rho', 'h', 'm', 'cs', 'p', 'e',
        'au', 'av', 'aw', 'arho', 'ae', 'am', 'ah', 'x0', 'y0', 'z0',
        'u0', 'v0', 'w0', 'rho0', 'e0', 'h0', 'div', 'dt_cfl',
        'grhox', 'grhoy', 'grhoz', 'dwdh', 'omega', 'converged',
        'alpha1', 'alpha10', 'aalpha1', 'alpha2', 'alpha20', 'aalpha2',
        'del2e'
    ]

    pa = get_particle_array(
        constants=constants, additional_props=required_props, **props
    )

    # set the intial smoothing length h0 to the particle smoothing
    # length. This can result in an annoying error in the density
    # iterations which require the h0 array
    pa.h0[:] = pa.h[:]

    pa.set_output_arrays(['x', 'y', 'u', 'v', 'rho', 'm', 'h', 'cs', 'p', 'e',
                          'au', 'av', 'ae', 'pid', 'gid', 'tag', 'dwdh',
                          'alpha1', 'alpha2'])

    return pa


def get_particle_array_swe(constants=None, **props):
    r"""**Return a particle array for the shallow water formulation**

    This sets the default properties to be::

        ['x', 'y', 'z', 'u', 'v', 'w', 'h', 'rho', 'arho' 'm', 'p', 'V', 'A',
        'cs', 'n', 'rho0', 'rho_prev_iter', 'rho_residual',
        'positive_rho_residual', 'summation_rho', 'dw', 'alpha', 'exp_lambda',
        'tv', 'tu', 'au', 'av', 'u_prev_step', 'v_prev_step', 'uh', 'vh',
        'dt_cfl', 'pa_to_split', 'Sfx', 'Sfy', 'psi', 'sum_Ak', 'u_parent',
        'v_parent', 'uh_parent', 'vh_parent', 'parent_idx', 'b', 'bx', 'by',
        'bxx', 'bxy', byy', 'closest_idx', 'is_merged_pa', 'merge',
        'dw_inner_reimann', 'u_inner_reimann', 'v_inner_reimann', 'shep_corr',
        'is_wall_boun_pa', 'dw_at_t', 'pa_out_of_domain', 'ob_pa_to_remove',
        'ob_pa_to_tag', 'pa_alpha_zero', 'fluid_pa_to_remove']

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """
    swe_props = ['A', 'cs', 'n', 'rho0', 'h0', 'rho_prev_iter', 'rho_residual',
                 'positive_rho_residual', 'summation_rho', 'dw', 'alpha',
                 'exp_lambda', 'tv', 'tu', 'u_prev_step', 'v_prev_step',
                 'uh', 'vh', 'dt_cfl', 'pa_to_split', 'Sfx', 'Sfy', 'V',
                 'psi', 'sum_Ak', 'arho', 'u_parent', 'v_parent', 'uh_parent',
                 'vh_parent', 'b', 'bx', 'by', 'bxx', 'bxy', 'byy', 'merge',
                 'is_merged_pa', 'dw_inner_reimann', 'u_inner_reimann',
                 'v_inner_reimann', 'shep_corr', 'is_wall_boun_pa', 'dw_at_t',
                 'pa_out_of_domain', 'ob_pa_to_remove', 'ob_pa_to_tag',
                 'pa_alpha_zero', 'fluid_pa_to_remove']

    consts = {}
    if constants:
        for key, val in constants.items():
            consts[key] = val

    pa = get_particle_array(
        constants=consts, additional_props=swe_props, **props
    )
    pa.add_property('parent_idx', type='int')
    pa.add_property('closest_idx', type='int')

    # default property arrays to save out.
    props = ['x', 'y', 'h', 'rho', 'h0', 'rho0', 'p', 'A', 'uh', 'vh', 'u',
             'v', 'm', 'tu', 'tv', 'dw', 'alpha', 'au', 'av', 'Sfx', 'Sfy',
             'n', 'cs', 'pid', 'gid', 'tag', 'b', 'bx', 'by', 'bxx', 'bxy',
             'byy']
    pa.set_output_arrays(props)

    return pa


def get_particles_info(particles):
    """Return the array information for a list of particles.

    Returns
    -------

    An OrderedDict containing the property information for a list of
    particles. This dict can be used for example to set-up dummy/empty
    particle arrays.

    """
    info = OrderedDict()
    for parray in particles:
        prop_info = {}
        for prop_name, prop in parray.properties.items():
            prop_info[prop_name] = {
                'name': prop_name, 'type': prop.get_c_type(),
                'default': parray.default_values[prop_name],
                'stride': parray.stride.get(prop_name, 1),
                'data': None}
        const_info = {}
        if parray.gpu is not None:
            parray.gpu.pull(*list(parray.constants.keys()))
        for c_name, value in parray.constants.items():
            const_info[c_name] = value.get_npy_array()
        info[parray.name] = dict(
            properties=prop_info, constants=const_info,
            output_property_arrays=parray.output_property_arrays,
            lb_props=parray.get_lb_props()
        )

    return info


def create_dummy_particles(info):
    """Returns a replica (empty) of a list of particles"""
    particles = []
    for name, pa_data in info.items():
        prop_dict = pa_data['properties']
        constants = pa_data['constants']
        lb_props = pa_data['lb_props']
        pa = ParticleArray(name=name, constants=constants, **prop_dict)
        pa.set_output_arrays(pa_data['output_property_arrays'])
        pa.set_lb_props(lb_props)
        particles.append(pa)

    return particles


def is_overloaded_method(method):
    """Returns True if the given method is overloaded from any of its bases.
    """
    method_name = method.__name__
    klass = method.__self__.__class__
    count = 0
    prev = None
    for base in klass.mro():
        if hasattr(base, method_name):
            method = getattr(base, method_name)
            if method != prev:
                prev = method
                count += 1
        if count > 1:
            break

    return count > 1



"""
Inlet boundary
"""
import numpy as np
from pysph.sph.bc.inlet_outlet_manager import InletBase


class Inlet(InletBase):
    def update(self, time, dt, stage):
        dest_pa = self.dest_pa
        inlet_pa = self.inlet_pa
        ghost_pa = self.ghost_pa

        dest_pa.uref[0] = 0.5 * (inlet_pa.uref[0] + dest_pa.uref[0])

        if not self._init:
            self.initialize()
            self._init = True
        if stage in self.active_stages:

            self.io_eval = self._create_io_eval()
            self.io_eval.update()
            self.io_eval.evaluate()

            io_id = inlet_pa.ioid
            cond = (io_id == 0)
            all_idx = np.where(cond)[0]
            inlet_pa.extract_particles(all_idx, dest_pa)

            # moving the moved particles back to the array beginning.
            inlet_pa.x[all_idx] += self.length * self.xn
            inlet_pa.y[all_idx] += self.length * self.yn
            inlet_pa.z[all_idx] += self.length * self.zn

            if ghost_pa:
                ghost_pa.x[all_idx] -= self.length * self.xn
                ghost_pa.y[all_idx] -= self.length * self.yn
                ghost_pa.z[all_idx] -= self.length * self.zn

            if self.callback is not None:
                self.callback(dest_pa, inlet_pa)



"""
Exact solution to Riemann problems.
"""

import numpy
from math import sqrt


def set_gamma(g):
    global gamma, gp1_2g, gm1_2g, gm1_gp1, gm1_2, gm1, gp1
    gamma = g
    gm1_2g = (gamma - 1.0) / (2.0 * gamma)
    gp1_2g = (gamma + 1.0) / (2.0 * gamma)
    gm1_gp1 = (gamma - 1.0) / (gamma + 1.0)
    gm1_2 = (gamma - 1.0) / 2.0
    gm1 = gamma - 1.0
    gp1 = gamma + 1.0


def solve(x_min=-0.5, x_max=0.5, x_0=0.0, t=0.1, p_l=1.0, p_r=0.1, rho_l=1.0,
          rho_r=0.125, u_l=0.0, u_r=0.0, N=101):

    r"""
    Parameters
    ----------
    x_min : float
        the leftmost point of domain
    x_max : float
        the rightmost point of domain
    x_0 : float
        the position of the diaphgram
    t : float
        total time of simulation
    p_l, u_l, rho_l : float
        pressure, velocity, density in the left region
    p_r, u_r, rho_r : float
        pressure, velocity, density in the right region
    N : int
        number of points under study

    The default arguments mentioned correspond to the Sod shock tube case.

    Notes
    -----
    The function returns the exact solution in the order of density, velocity,
    pressure, energy and x-coordinates of the points under study.

    References
    ----------
    .. E.F. Toro, Riemann Solvers and Numerical Methods for Fluid Dynamics,
        Springer (2009), Chapter 4, pp. 115-138

    """
    c_l = sqrt(gamma * p_l / rho_l)
    c_r = sqrt(gamma * p_r / rho_r)

    try:
        import scipy
        print("Using fsolve to solve the non-linear equation")
        p_star, u_star = star_pu_fsolve(rho_l, u_l, p_l, c_l,
                                        rho_r, u_r, p_r, c_r)
    except ImportError:
        print("Using Newton-Raphson method to solve the non-linear equation")
        p_star, u_star = star_pu_newton_raphson(rho_l, u_l, p_l, c_l,
                                                rho_r, u_r, p_r, c_r)

    # check if the discontinuity is inside the domain
    msg = "discontinuity not in domain"
    assert x_0 >= x_min and x_0 <= x_max, msg

    # transform domain according to initial discontinuity
    x_min = x_min - x_0
    x_max = x_max - x_0

    print('p_star=' + str(p_star))
    print('u_star=' + str(u_star))
    x = numpy.linspace(x_min, x_max, N)
    density = []
    pressure = []
    velocity = []
    energy = []
    for i in range(0, N):
        s = x[i] / t
        rho, u, p = complete_solution(rho_l, u_l, p_l, c_l, rho_r, u_r,
                                      p_r, c_r, p_star, u_star, s)
        density.append(rho)
        velocity.append(u)
        pressure.append(p)
        energy.append(p / (gm1 * rho))
    # transform the domain back to original coordinates
    x = x + x_0
    return tuple(map(numpy.asarray, [density, velocity, pressure, energy, x]))


def _flux_fsolve(pressure, rho1, c1, p1):
    if pressure <= p1:  # Rarefaction
        return lambda p: (2 / gm1) * c1 * ((p / p1)**gm1_2g - 1.0)
    else:  # Shock
        return lambda p: (
            (p - p1) * sqrt(((2 / gp1) / rho1) / ((gm1_gp1 * p1) + p))
        )


def star_pu_fsolve(rho_l, u_l, p_l, c_l, rho_r, u_r, p_r, c_r):
    p_min = min(p_l, p_r)
    p_max = max(p_l, p_r)
    f_min = _flux_fsolve(p_min, rho_l, c_l, p_l)(p_min) + \
        _flux_fsolve(p_min, rho_r, c_r, p_r)(p_min) + u_r - u_l
    f_max = _flux_fsolve(p_max, rho_l, c_l, p_l)(p_max) + \
        _flux_fsolve(p_max, rho_r, c_r, p_r)(p_max) + u_r - u_l
    if (f_min > 0 and f_max > 0):
        p_guess = 0.5 * (0 + p_min)
        p_star, u_star = _star_pu(rho_l, u_l, p_l, c_l, rho_r,
                                  u_r, p_r, c_r, p_guess)
    elif(f_min <= 0 and f_max >= 0):
        p_guess = (p_l + p_r) * 0.5
        p_star, u_star = _star_pu(rho_l, u_l, p_l, c_l, rho_r,
                                  u_r, p_r, c_r, p_guess)
    else:
        p_guess = 2 * p_max
        p_star, u_star = _star_pu(rho_l, u_l, p_l, c_l, rho_r,
                                  u_r, p_r, c_r, p_guess)
    return p_star, u_star


def _star_pu(rho_l, u_l, p_l, c_l, rho_r, u_r, p_r, c_r, p_guess):
    """Computes the pressure and velocity in the star region using fsolve
       from scipy module"""
    fl = _flux_fsolve(p_guess, rho_l, c_l, p_l)
    fr = _flux_fsolve(p_guess, rho_r, c_r, p_r)
    f = lambda p: fl(p) + fr(p) + u_r - u_l
    from scipy.optimize import fsolve
    p_star = fsolve(f, 0.0)[0]
    u_star = (
        0.5 * (u_l + u_r + _flux_fsolve(p_star, rho_r, c_r, p_r)(p_star) -
               _flux_fsolve(p_star, rho_l, c_l, p_l)(p_star))
    )
    return p_star, u_star


def _flux_newton(pressure, rho1, c1, p1):
    if pressure <= p1:  # Rarefaction
        flux = (2 / gm1) * c1 * ((pressure / p1)**gm1_2g - 1.0)
        flux_derivative = (1.0 / (rho1 * c1)) * \
            (pressure / p1)**(-gp1_2g)
        return flux, flux_derivative
    else:  # Shock
        flux = (
            (pressure - p1) * sqrt(((2 / gp1) / rho1) /
                                   ((gm1_gp1 * p1) + pressure))
        )
        flux_derivative = (
            (1.0 - 0.5 * (pressure - p1) / ((gm1_gp1 * p1) + pressure)) *
            sqrt(((2 / gp1) / rho1) / ((gm1_gp1 * p1) + pressure))
        )
        return flux, flux_derivative


def star_pu_newton_raphson(rho_l, u_l, p_l, c_l, rho_r, u_r, p_r, c_r):
    tol_pre = 1.0e-06
    nr_iter = 20
    p_start = _compute_guess_p(rho_l, u_l, p_l, c_l, rho_r, u_r, p_r, c_r)
    p_old = p_start
    u_diff = u_r - u_l
    for i in range(nr_iter):
        fL, fLd = _flux_newton(p_old, rho_l, c_l, p_l)
        fR, fRd = _flux_newton(p_old, rho_r, c_r, p_r)
        p = p_old - (fL + fR + u_diff) / (fLd + fRd)
        change = 2.0 * abs((p - p_old) / (p + p_old))
        if change <= tol_pre:
            break
        if p < 0.0:
            p = tol_pre
        p_old = p
    u = 0.5 * (u_l + u_r + fR - fL)
    return p, u


def _compute_guess_p(rho_l, u_l, p_l, c_l, rho_r, u_r, p_r, c_r):
    """ Computes the initial guess for pressure.
    References
    ----------
    E.F. Toro, Riemann Solvers and Numerical Methods for Fluid Dynamics
    Springer (2009), Chapter 9, pp. 297-306
    """
    quser = 2.0
    p_linearized = 0.5 * (p_l + p_r) + 0.5 * (u_l - u_r) * \
        0.25 * (rho_l + rho_r) * (c_l + c_r)
    p_linearized = max(0.0, p_linearized)
    p_min = min(p_l, p_r)
    p_max = max(p_l, p_r)
    qmax = p_max / p_min
    if(
        qmax <= quser and (p_min <= p_linearized and
                           p_linearized <= p_max)
    ):
        """A Primitive Variable Riemann Solver (PMRS)"""
        return p_linearized
    else:
        """A Two-Rarefaction Riemann Solver (TRRS)"""
        if p_linearized < p_min:
            p_lr = (p_l / p_r)**gm1_2g
            u_linearized = (p_lr * u_l / c_l + u_r / c_r + (2 / gm1) *
                            (p_lr - 1.0)) / (p_lr / c_l + 1.0 / c_r)
            return (
                0.5 * (p_l * (1.0 + gm1_2 * (u_l - u_linearized) /
                              c_l)**(1.0 / gm1_2g) +
                       p_r * (1.0 + gm1_2 * (u_linearized - u_r) / c_r) **
                       (1.0 / gm1_2g))
            )
        else:
            """A Two-Shock Riemann Solver (TSRS)"""
            gL = sqrt(((2 / gp1) / rho_l) /
                      (gm1_gp1 * p_l + p_linearized))
            gR = sqrt(((2 / gp1) / rho_r) /
                      (gm1_gp1 * p_r + p_linearized))
            return (gL * p_l + gR * p_r - (u_r - u_l)) / (gL + gR)


def complete_solution(rho_l, u_l, p_l, c_l, rho_r, u_r, p_r, c_r, p_star,
                      u_star, s):
    if s <= u_star:
        rho, u, p = left_contact(rho_l, u_l, p_l, c_l, p_star, u_star, s)
    else:
        rho, u, p = right_contact(rho_r, u_r, p_r, c_r, p_star, u_star, s)
    return rho, u, p


def left_contact(rho_l, u_l, p_l, c_l, p_star, u_star, s):
    if p_star <= p_l:
        rho, u, p = left_rarefaction(rho_l, u_l, p_l, c_l, p_star, u_star, s)
    else:
        rho, u, p = left_shock(rho_l, u_l, p_l, c_l, p_star, u_star, s)
    return rho, u, p


def left_rarefaction(rho_l, u_l, p_l, c_l, p_star, u_star, s):
    s_head = u_l - c_l
    s_tail = u_star - c_l * (p_star / p_l)**gm1_2g
    if s <= s_head:
        rho, u, p = rho_l, u_l, p_l
    elif (s > s_head and s <= s_tail):
        u = (2 / gp1) * (c_l + gm1_2 * u_l + s)
        c = (2 / gp1) * (c_l + gm1_2 * (u_l - s))
        rho = rho_l * (c / c_l)**(2 / gm1)
        p = p_l * (c / c_l)**(1.0 / gm1_2g)
    else:
        rho = rho_l * (p_star / p_l)**(1.0 / gamma)
        u = u_star
        p = p_star
    return rho, u, p


def left_shock(rho_l, u_l, p_l, c_l, p_star, u_star, s):
    sL = u_l - c_l * sqrt(gp1_2g * (p_star / p_l) + gm1_2g)
    if s <= sL:
        rho, u, p = rho_l, u_l, p_l
    else:
        rho_1 = rho_l * ((p_star / p_l) + gm1_gp1) / \
            ((p_star / p_l) * gm1_gp1 + 1.0)
        rho, u, p = rho_1, u_star, p_star
    return rho, u, p


def right_contact(rho_r, u_r, p_r, c_r, p_star, u_star, s):
    if p_star > p_r:
        rho, u, p = right_shock(rho_r, u_r, p_r, c_r, p_star, u_star, s)
    else:
        rho, u, p = right_rarefaction(rho_r, u_r, p_r, c_r, p_star, u_star, s)
    return rho, u, p


def right_rarefaction(rho_r, u_r, p_r, c_r, p_star, u_star, s):
    s_head = u_r + c_r
    s_tail = u_star + c_r * (p_star / p_r)**gm1_2g
    if s >= s_head:
        rho, u, p = rho_r, u_r, p_r
    elif (s < s_head and s > s_tail):
        u = (2 / gp1) * (-c_r + gm1_2 * u_r + s)
        c = (2 / gp1) * (c_r - gm1_2 * (u_r - s))
        rho = rho_r * (c / c_r)**(2 / gm1)
        p = p_r * (c / c_r)**(1.0 / gm1_2g)
    else:
        rho = rho_r * (p_star / p_r)**(1.0 / gamma)
        u = u_star
        p = p_star
    return rho, u, p


def right_shock(rho_r, u_r, p_r, c_r, p_star, u_star, s):
    sR = u_r + c_r * sqrt(gp1_2g * (p_star / p_r) + gm1_2g)
    if s >= sR:
        rho, u, p = rho_r, u_r, p_r
    else:
        rho_1 = rho_r * ((p_star / p_r) + gm1_gp1) / \
            ((p_star / p_r) * gm1_gp1 + 1.0)
        rho, u, p = rho_1, u_star, p_star
    return rho, u, p


if __name__ == '__main__':
    set_gamma(1.4)
    solve()



"""Rayleigh-Taylor instability problem. (16 hours)
"""

# numpy
import numpy as np

# PySPH imports
from pysph.base.utils import get_particle_array
from pysph.solver.solver import Solver
from pysph.solver.application import Application

from pysph.sph.scheme import TVFScheme


# domain and reference values
gy = -1.0
Lx = 1.0
Ly = 2.0
Re = 420
Vmax = np.sqrt(0.5 * Ly * abs(gy))
nu = Vmax * Ly / Re

# density for the two phases
rho1 = 1.8
rho2 = 1.0

# speed of sound and reference pressure
Fr = 0.01
c0 = Vmax / Fr
p1 = c0**2 * rho1
p2 = c0**2 * rho2

# Numerical setup
nx = 50
dx = Lx / nx
ghost_extent = 5 * dx
hdx = 1.2

# adaptive time steps
h0 = hdx * dx
dt_cfl = 0.25 * h0 / (c0 + Vmax)
dt_viscous = 0.125 * h0**2 / nu
dt_force = 0.25 * np.sqrt(h0 / abs(gy))

tf = 25
dt = 0.5 * min(dt_cfl, dt_viscous, dt_force)


class RayleighTaylor(Application):
    def create_particles(self):
        # create all the particles
        _x = np.arange(-ghost_extent - dx / 2, Lx + ghost_extent + dx / 2, dx)
        _y = np.arange(-ghost_extent - dx / 2, Ly + ghost_extent + dx / 2, dx)
        x, y = np.meshgrid(_x, _y)
        x = x.ravel()
        y = y.ravel()

        # sort out the fluid and the solid
        indices = []
        for i in range(x.size):
            if ((x[i] > 0.0) and (x[i] < Lx)):
                if ((y[i] > 0.0) and (y[i] < Ly)):
                    indices.append(i)

        # create the arrays
        solid = get_particle_array(name='solid', x=x, y=y)

        # remove the fluid particles from the solid
        fluid = solid.extract_particles(indices)
        fluid.set_name('fluid')
        solid.remove_particles(indices)

        # sort out the two fluid phases
        indices = []
        for i in range(fluid.get_number_of_particles()):
            if fluid.y[i] > 1 - 0.15 * np.sin(2 * np.pi * fluid.x[i]):
                indices.append(i)

        fluid1 = fluid.extract_particles(indices)
        fluid1.set_name('fluid1')
        fluid2 = fluid
        fluid2.set_name('fluid2')
        fluid2.remove_particles(indices)

        fluid1.rho[:] = rho1
        fluid2.rho[:] = rho2

        print("Rayleigh Taylor Instability problem :: Re = %d, nfluid = %d, nsolid=%d, dt = %g" % (
            Re, fluid1.get_number_of_particles() + fluid2.get_number_of_particles(),
            solid.get_number_of_particles(), dt))

        # add requisite properties to the arrays:
        self.scheme.setup_properties([fluid1, fluid2, solid])

        # setup the particle properties
        volume = dx * dx

        # mass is set to get the reference density of each phase
        fluid1.m[:] = volume * rho1
        fluid2.m[:] = volume * rho2

        # volume is set as dx^2
        fluid1.V[:] = 1. / volume
        fluid2.V[:] = 1. / volume
        solid.V[:] = 1. / volume

        # smoothing lengths
        fluid1.h[:] = hdx * dx
        fluid2.h[:] = hdx * dx
        solid.h[:] = hdx * dx

        # return the arrays
        return [fluid1, fluid2, solid]

    def create_scheme(self):
        s = TVFScheme(
            ['fluid1', 'fluid2'], ['solid'], dim=2, rho0=rho1, c0=c0, nu=nu,
            p0=p1, pb=p1, h0=dx * hdx, gy=gy
        )
        s.configure_solver(tf=tf, dt=dt, pfreq=500)
        return s

    def create_equations(self):
        # This is an ugly hack to support different densities for fluids.
        # What we should really do is set rho0 as a fluid constant and rewrite
        # the equations to use that, then once the fluid properties are
        # defined, this will just work.
        equations = super(RayleighTaylor, self).create_equations()
        from pysph.sph.equation import Group

        def process_term(x):
            if hasattr(x, 'rho0'):
                if x.dest == 'fluid1' or x.sources == ['fluid1']:
                    x.rho0 = rho1
                elif x.dest == 'fluid2' or x.sources == ['fluid2']:
                    x.rho0 = rho2
            if hasattr(x, 'p0'):
                if x.dest == 'fluid1':
                    x.p0 = p1
                elif x.dest == 'fluid2':
                    x.p0 = p2

        for eq in equations:
            if isinstance(eq, Group):
                for eq in eq.equations:
                    process_term(eq)
            else:
                process_term(eq)

        return equations


if __name__ == '__main__':
    app = RayleighTaylor()
    app.run()



from __future__ import print_function
import numpy as np

from pysph.base.nnps_base import DomainManagerBase

from compyle.config import get_config
from compyle.array import Array, get_backend
from compyle.parallel import Elementwise, Reduction, Scan
from compyle.types import annotate, declare

from pytools import memoize_method


class GPUDomainManager(DomainManagerBase):
    def __init__(self, xmin=-1000., xmax=1000., ymin=0.,
                 ymax=0., zmin=0., zmax=0.,
                 periodic_in_x=False, periodic_in_y=False,
                 periodic_in_z=False, n_layers=2.0, backend=None, props=None,
                 mirror_in_x=False, mirror_in_y=False, mirror_in_z=False):
        """Constructor"""
        DomainManagerBase.__init__(self, xmin=xmin, xmax=xmax,
                                   ymin=ymin, ymax=ymax, zmin=zmin, zmax=zmax,
                                   periodic_in_x=periodic_in_x,
                                   periodic_in_y=periodic_in_y,
                                   periodic_in_z=periodic_in_z,
                                   n_layers=n_layers, props=props)

        self.use_double = get_config().use_double
        self.dtype = np.float64 if self.use_double else np.float32

        self.dtype_max = np.finfo(self.dtype).max
        self.backend = get_backend(backend)

        self.ghosts = None

    def update(self):
        """General method that is called before NNPS can bin particles.

        This method is responsible for the computation of cell sizes
        and creation of any ghost particles for periodic or wall
        boundary conditions.

        """
        # compute the cell sizes
        self.compute_cell_size_for_binning()

        # Periodicity is handled by adjusting particles according to a
        # given cubic domain box. In parallel, it is expected that the
        # appropriate parallel NNPS is responsible for the creation of
        # ghost particles.
        if self.is_periodic and not self.in_parallel:
            # remove periodic ghost particles from a previous step
            self._remove_ghosts()

            # box-wrap current particles for periodicity
            self._box_wrap_periodic()

            # create new periodic ghosts
            self._create_ghosts_periodic()

    def _compute_cell_size_for_binning(self):
        """Compute the cell size for the binning.

        The cell size is chosen as the kernel radius scale times the
        maximum smoothing length in the local processor. For parallel
        runs, we would need to communicate the maximum 'h' on all
        processors to decide on the appropriate binning size.

        """
        _hmax, hmax = -1.0, -1.0
        _hmin, hmin = self.dtype_max, self.dtype_max

        for pa_wrapper in self.pa_wrappers:
            h = pa_wrapper.pa.gpu.get_device_array('h')
            pa_wrapper.pa.gpu.update_minmax_cl(['h'])

            _hmax = h.maximum
            _hmin = h.minimum

            if _hmax > hmax:
                hmax = _hmax
            if _hmin < hmin:
                hmin = _hmin

        cell_size = self.radius_scale * hmax
        self.hmin = self.radius_scale * hmin

        if cell_size < 1e-6:
            cell_size = 1.0

        self.cell_size = cell_size

        # set the cell size for the DomainManager
        self.set_cell_size(cell_size)

    @memoize_method
    def _get_box_wrap_kernel(self):
        @annotate
        def box_wrap(i, x, y, z, xmin, ymin, zmin, xmax, ymax, zmax,
                     xtranslate, ytranslate, ztranslate,
                     periodic_in_x, periodic_in_y, periodic_in_z):
            if periodic_in_x:
                if x[i] < xmin:
                    x[i] = x[i] + xtranslate
                if x[i] > xmax:
                    x[i] = x[i] - xtranslate

            if periodic_in_y:
                if y[i] < ymin:
                    y[i] = y[i] + ytranslate
                if y[i] > ymax:
                    y[i] = y[i] - ytranslate

            if periodic_in_z:
                if z[i] < zmin:
                    z[i] = z[i] + ztranslate
                if z[i] > zmax:
                    z[i] = z[i] - ztranslate

        return Elementwise(box_wrap, backend=self.backend)

    ###########################CHANGE FROM HERE###############################

    def _box_wrap_periodic(self):
        """Box-wrap particles for periodicity

        The periodic domain is a rectangular box defined by minimum
        and maximum values in each coordinate direction. These values
        are used in turn to define translation values used to box-wrap
        particles that cross a periodic boundary.

        The periodic domain is specified using the DomainManager object

        """
        # minimum and maximum values of the domain
        xmin, xmax = self.xmin, self.xmax
        ymin, ymax = self.ymin, self.ymax
        zmin, zmax = self.zmin, self.zmax

        # translations along each coordinate direction
        xtranslate = self.xtranslate
        ytranslate = self.ytranslate
        ztranslate = self.ztranslate

        # periodicity flags for NNPS
        periodic_in_x = self.periodic_in_x
        periodic_in_y = self.periodic_in_y
        periodic_in_z = self.periodic_in_z

        box_wrap_knl = self._get_box_wrap_kernel()

        # iterate over each array and mark for translation
        for pa_wrapper in self.pa_wrappers:
            x = pa_wrapper.pa.gpu.x
            y = pa_wrapper.pa.gpu.y
            z = pa_wrapper.pa.gpu.z

            box_wrap_knl(x, y, z, xmin, ymin, zmin, xmax, ymax, zmax,
                         xtranslate, ytranslate, ztranslate,
                         periodic_in_x, periodic_in_y, periodic_in_z)

    @memoize_method
    def _get_ghosts_reduction_kernel(self):
        @annotate
        def map_func(i, x, y, z, xmin, ymin, zmin, xmax, ymax, zmax, cell_size,
                     periodic_in_x, periodic_in_y, periodic_in_z):
            x_copies, y_copies, z_copies = declare('int', 3)

            x_copies = 1
            y_copies = 1
            z_copies = 1

            if periodic_in_x:
                if (x[i] - xmin) <= cell_size:
                    x_copies += 1
                if (xmax - x[i]) <= cell_size:
                    x_copies += 1

            if periodic_in_y:
                if (y[i] - ymin) <= cell_size:
                    y_copies += 1
                if (ymax - y[i]) <= cell_size:
                    y_copies += 1

            if periodic_in_z:
                if (z[i] - zmin) <= cell_size:
                    z_copies += 1
                if (zmax - z[i]) <= cell_size:
                    z_copies += 1

            return x_copies * y_copies * z_copies - 1

        return Reduction('a+b', map_func=map_func, dtype_out=np.int32,
                         backend=self.backend)

    @memoize_method
    def _get_ghosts_scan_kernel(self):
        @annotate
        def inp_fill_ghosts(i, periodic_in_x, periodic_in_y, periodic_in_z,
                            x, y, z, xmin, ymin, zmin, xmax, ymax, zmax,
                            cell_size):
            x_copies, y_copies, z_copies = declare('int', 3)

            x_copies = 1
            y_copies = 1
            z_copies = 1

            if periodic_in_x:
                if (x[i] - xmin) <= cell_size:
                    x_copies += 1
                if (xmax - x[i]) <= cell_size:
                    x_copies += 1

            if periodic_in_y:
                if (y[i] - ymin) <= cell_size:
                    y_copies += 1
                if (ymax - y[i]) <= cell_size:
                    y_copies += 1

            if periodic_in_z:
                if (z[i] - zmin) <= cell_size:
                    z_copies += 1
                if (zmax - z[i]) <= cell_size:
                    z_copies += 1

            return x_copies * y_copies * z_copies - 1

        @annotate
        def out_fill_ghosts(i, item, prev_item, periodic_in_x,
                            periodic_in_y, periodic_in_z, x, y, z,
                            xmin, ymin, zmin, xmax, ymax, zmax, cell_size,
                            masks, indices):
            xleft, yleft, zleft = declare('int', 3)
            xright, yright, zright = declare('int', 3)

            xleft = 0
            yleft = 0
            zleft = 0

            xright = 0
            yright = 0
            zright = 0

            if periodic_in_x:
                if (x[i] - xmin) <= cell_size:
                    xright = 1
                if (xmax - x[i]) <= cell_size:
                    xleft = -1

            if periodic_in_y:
                if (y[i] - ymin) <= cell_size:
                    yright = 1
                if (ymax - y[i]) <= cell_size:
                    yleft = -1

            if periodic_in_z:
                if (z[i] - zmin) <= cell_size:
                    zright = 1
                if (zmax - z[i]) <= cell_size:
                    zleft = -1

            xp, yp, zp = declare('int', 3)
            idx, mask = declare('int', 2)

            idx = prev_item

            for xp in range(-1, 2):
                if xp != 0 and ((xleft == 0 and xright == 0) or
                                (xp != xleft and xp != xright)):
                    continue
                for yp in range(-1, 2):
                    if yp != 0 and ((yleft == 0 and yright == 0) or
                                    (yp != yleft and yp != yright)):
                        continue
                    for zp in range(-1, 2):
                        if zp != 0 and ((zleft == 0 and zright == 0) or
                                        (zp != zleft and zp != zright)):
                            continue
                        if xp == 0 and yp == 0 and zp == 0:
                            continue
                        mask = (xp + 1) * 9 + (yp + 1) * 3 + (zp + 1)
                        masks[idx] = mask
                        indices[idx] = i
                        idx += 1

        return Scan(inp_fill_ghosts, out_fill_ghosts, 'a+b',
                    dtype=np.int32, backend=self.backend)

    @memoize_method
    def _get_translate_kernel(self):
        @annotate
        def translate(i, x, y, z, tag, xtranslate, ytranslate,
                      ztranslate, masks):
            xmask, ymask, zmask, mask = declare('int', 4)
            mask = masks[i]

            zmask = mask % 3
            mask /= 3
            ymask = mask % 3
            mask /= 3
            xmask = mask % 3

            x[i] = x[i] + (xmask - 1) * xtranslate
            y[i] = y[i] + (ymask - 1) * ytranslate
            z[i] = z[i] + (zmask - 1) * ztranslate

            tag[i] = 2

        return Elementwise(translate, backend=self.backend)

    def _create_ghosts_periodic(self):
        """Identify boundary particles and create images.

        We need to find all particles that are within a specified
        distance from the boundaries and place image copies on the
        other side of the boundary. Corner reflections need to be
        accounted for when using domains with multiple periodicity.

        The periodic domain is specified using the DomainManager object

        """
        copy_props = self.copy_props
        pa_wrappers = self.pa_wrappers
        narrays = self.narrays

        # cell size used to check for periodic ghosts. For summation density
        # like operations, we need to create two layers of ghost images, this
        # is configurable via the n_layers argument to the constructor.
        cell_size = self.n_layers * self.cell_size

        # periodic domain values
        xmin, xmax = self.xmin, self.xmax
        ymin, ymax = self.ymin, self.ymax
        zmin, zmax = self.zmin, self.zmax

        xtranslate = self.xtranslate
        ytranslate = self.ytranslate
        ztranslate = self.ztranslate

        # periodicity flags
        periodic_in_x = self.periodic_in_x
        periodic_in_y = self.periodic_in_y
        periodic_in_z = self.periodic_in_z

        reduce_knl = self._get_ghosts_reduction_kernel()
        scan_knl = self._get_ghosts_scan_kernel()
        translate_knl = self._get_translate_kernel()

        if not self.ghosts:
            self.ghosts = [paw.pa.empty_clone(props=copy_props[i])
                           for i, paw in enumerate(pa_wrappers)]
        else:
            for ghost_pa in self.ghosts:
                ghost_pa.resize(0)
            for i in range(narrays):
                self.ghosts[i].ensure_properties(
                    pa_wrappers[i].pa, props=copy_props[i]
                )

        for i, pa_wrapper in enumerate(self.pa_wrappers):
            ghost_pa = self.ghosts[i]

            x = pa_wrapper.pa.gpu.x
            y = pa_wrapper.pa.gpu.y
            z = pa_wrapper.pa.gpu.z

            num_extra_particles = reduce_knl(x, y, z, xmin, ymin, zmin, xmax,
                                             ymax, zmax, cell_size,
                                             periodic_in_x, periodic_in_y,
                                             periodic_in_z)

            num_extra_particles = int(num_extra_particles)

            indices = Array(np.int32, n=num_extra_particles)
            masks = Array(np.int32, n=num_extra_particles)

            scan_knl(periodic_in_x=periodic_in_x, periodic_in_y=periodic_in_y,
                     periodic_in_z=periodic_in_z, x=x, y=y, z=z, xmin=xmin,
                     ymin=ymin, zmin=zmin, xmax=xmax, ymax=ymax, zmax=zmax,
                     cell_size=cell_size, masks=masks, indices=indices)

            pa_wrapper.pa.extract_particles(
                indices, ghost_pa, align=False, props=copy_props[i]
            )

            translate_knl(ghost_pa.gpu.x, ghost_pa.gpu.y, ghost_pa.gpu.z,
                          ghost_pa.gpu.tag, xtranslate, ytranslate,
                          ztranslate, masks)

            pa_wrapper.pa.append_parray(ghost_pa, align=False)



"""Basic code for the templated integrators.

Currently we only support two-step integrators.

These classes are used to generate the code for the actual integrators
from the `sph_eval` module.
"""

from numpy import sqrt
import numpy as np

from compyle.profile import profile_ctx, profile
# Local imports.
from .integrator_step import IntegratorStep


###############################################################################
# `Integrator` class
###############################################################################
class Integrator(object):
    r"""Generic class for multi-step integrators in PySPH for a system of
    ODES of the form :math:`\frac{dy}{dt} = F(y)`.
    """

    def __init__(self, **kw):
        """Pass fluid names and suitable `IntegratorStep` instances.

        For example::

            >>> integrator = Integrator(fluid=WCSPHStep(), solid=WCSPHStep())

        where "fluid" and "solid" are the names of the particle arrays.
        """
        for array_name, integrator_step in kw.items():
            if not isinstance(integrator_step, IntegratorStep):
                msg = ('Stepper %s must be an instance of '
                       'IntegratorStep' % (integrator_step))
                raise ValueError(msg)

        self.steppers = kw
        self.parallel_manager = None
        self.nnps = None
        self.acceleration_evals = None
        # This is set later when the underlying compiled integrator is created
        # by the SPHCompiler.
        self.c_integrator = None
        self._has_dt_adapt = None
        self.fixed_h = False

    def __repr__(self):
        name = self.__class__.__name__
        s = self.steppers
        args = ', '.join(['%s=%s' % (k, s[k]) for k in s])
        return '%s(%s)' % (name, args)

    def _my_max(self, x):
        if len(x) > 0:
            return np.max(x)
        else:
            return -1.0

    def _get_dt_adapt_factors(self):
        a_eval = self.acceleration_evals[0]
        factors = [-1.0, -1.0, -1.0]
        for pa in a_eval.particle_arrays:
            prop_names = []
            for i, name in enumerate(('dt_cfl', 'dt_force', 'dt_visc')):
                if name in pa.properties:
                    if pa.gpu:
                        prop_names.append(name)
                    else:
                        max_val = self._my_max(pa.get(name))
                        factors[i] = max(factors[i], max_val)
            if pa.gpu:
                pa.gpu.update_minmax_cl(prop_names, only_max=True)
                for i, name in enumerate(('dt_cfl', 'dt_force', 'dt_visc')):
                    if name in pa.properties:
                        max_val = getattr(pa.gpu, name).maximum
                        factors[i] = max(factors[i], max_val)
        cfl_f, force_f, visc_f = factors
        return cfl_f, force_f, visc_f

    def _get_explicit_dt_adapt(self):
        """Checks if the user is defining a 'dt_adapt' property where the
        timestep is directly specified.

        This returns None if no such parameter is found, else it returns the
        allowed timestep.
        """
        a_eval = self.acceleration_evals[0]
        if self._has_dt_adapt is None:
            self._has_dt_adapt = any(
                'dt_adapt' in pa.properties for pa in a_eval.particle_arrays
            )
        if self._has_dt_adapt:
            dt_min = np.inf
            for pa in a_eval.particle_arrays:
                if 'dt_adapt' in pa.properties:
                    if pa.gpu is not None:
                        if pa.gpu.get_number_of_particles() > 0:
                            from compyle.array import minimum
                            min_val = minimum(pa.gpu.dt_adapt)
                        else:
                            min_val = np.inf
                    else:
                        if pa.get_number_of_particles() > 0:
                            min_val = np.min(pa.dt_adapt)
                        else:
                            min_val = np.inf
                    dt_min = min(dt_min, min_val)

            if dt_min > 0.0:
                return dt_min
            else:
                return None
        else:
            return None

    ##########################################################################
    # Public interface.
    ##########################################################################
    def set_acceleration_evals(self, a_evals):
        '''Set the acceleration evaluators.

        This must be done before the integrator is used.

        If you are using the SPHCompiler, it automatically calls this method.

        '''
        if isinstance(a_evals, (list, tuple)):
            self.acceleration_evals = a_evals
        else:
            self.acceleration_evals = [a_evals]

    def set_fixed_h(self, fixed_h):
        # compute h_minimum once for constant smoothing lengths
        if fixed_h:
            self.compute_h_minimum()

        self.fixed_h = fixed_h

    def set_nnps(self, nnps):
        self.nnps = nnps
        self.c_integrator.set_nnps(nnps)

    def compute_h_minimum(self):
        a_eval = self.acceleration_evals[0]

        hmin = 1.0
        for pa in a_eval.particle_arrays:
            if pa.gpu:
                h = pa.gpu.get_device_array('h')
            else:
                h = pa.get_carray('h')

            if h.minimum < hmin:
                hmin = h.minimum

        self.h_minimum = hmin

    def compute_time_step(self, dt, cfl):
        """If there are any adaptive timestep constraints, the appropriate
        timestep is returned, else None is returned.
        """
        dt_adapt = self._get_explicit_dt_adapt()
        if dt_adapt is not None:
            return dt_adapt

        dt_cfl_fac, dt_force_fac, dt_visc_fac = self._get_dt_adapt_factors()

        # iterate over particles and find hmin if using variable h
        if not self.fixed_h:
            self.compute_h_minimum()

        hmin = self.h_minimum

        # default time steps set to some large value
        dt_cfl = dt_force = dt_viscous = np.inf

        # stable time step based on courant condition
        if dt_cfl_fac > 0:
            dt_cfl = hmin/dt_cfl_fac

        # stable time step based on force criterion
        if dt_force_fac > 0:
            dt_force = sqrt(hmin/sqrt(dt_force_fac))

        # stable time step based on viscous condition
        if dt_visc_fac > 0:
            dt_viscous = hmin/dt_visc_fac

        # minimum of all three
        dt_min = min(dt_cfl, dt_force, dt_viscous)

        # return the computed time steps. If dt factors aren't
        # defined, the default dt is returned
        if dt_min <= 0.0 or np.isinf(dt_min):
            return None
        else:
            return cfl*dt_min

    def one_timestep(self, t, dt):
        """User written function that actually does one timestep.

        This function is used in the high-performance Cython implementation.
        The assumptions one may make are the following:

            - t and dt are passed.

            - the following methods are available:

                - self.initialize()

                - self.stage1(), self.stage2() etc. depending on the number of
                  stages available.

                - self.compute_accelerations(index=0, update_nnps=True)
                - self.do_post_stage(stage_dt, stage_count_from_1)
                - self.update_domain()

        Please see any of the concrete implementations of the Integrator class
        to study.  By default the Integrator implements a
        predict-evaluate-correct method, the same as PECIntegrator.

        """
        self.initialize()

        # Predict
        self.stage1()
        self.update_domain()

        # Call any post-stage functions.
        self.do_post_stage(0.5*dt, 1)

        self.compute_accelerations()

        # Correct
        self.stage2()
        self.update_domain()

        # Call any post-stage functions.
        self.do_post_stage(dt, 2)

    def set_compiled_object(self, c_integrator):
        """Set the high-performance compiled object to call internally.
        """
        self.c_integrator = c_integrator

    def set_parallel_manager(self, pm):
        self.parallel_manager = pm
        self.c_integrator.set_parallel_manager(pm)

    def set_post_stage_callback(self, callback):
        """This callback is called when the particles are moved, i.e
        one stage of the integration is done.

        This callback is passed the current time value, the timestep and the
        stage.

        The current time value is  t + stage_dt, for example this would be
        0.5*dt for a two stage predictor corrector integrator.

        """
        self.c_integrator.set_post_stage_callback(callback)

    def step(self, time, dt):
        """This function is called by the solver.

        To implement the integration step please override the
        ``one_timestep`` method.
        """
        self.c_integrator.step(time, dt)

    def compute_accelerations(self, index=0, update_nnps=True):
        if update_nnps:
            # update NNPS since particles have moved
            if self.parallel_manager:
                self.parallel_manager.update()
            with profile_ctx('nnps.update'):
                self.nnps.update()

        # Evaluate
        c_integrator = self.c_integrator
        a_eval = self.acceleration_evals[index]
        with profile_ctx('acceleration_eval_%d' % index):
            a_eval.compute(c_integrator.t, c_integrator.dt)

    @profile
    def initial_acceleration(self, t, dt):
        """Compute the initial accelerations if needed before the iterations start.

        The default implementation only does this for the first acceleration
        evaluator. So if you have multiple evaluators, you must override this
        method in a subclass.

        """
        self.acceleration_evals[0].compute(t, dt)

    @profile
    def update_domain(self):
        """Update the domain of the simulation.

        This is to be called when particles move so the ghost particles
        (periodicity, mirror boundary conditions) can be reset. Further, this
        also recalculates the appropriate cell size based on the particle
        kernel radius, `h`. This should be called explicitly when desired but
        usually this is done when the particles are moved or the `h` is
        changed.

        The integrator should explicitly call this when needed in the
        `one_timestep` method.
        """
        self.nnps.update_domain()


###############################################################################
# `EulerIntegrator` class
###############################################################################
class EulerIntegrator(Integrator):
    def one_timestep(self, t, dt):
        self.compute_accelerations()
        self.stage1()
        self.update_domain()
        self.do_post_stage(dt, 1)


###############################################################################
# `PECIntegrator` class
###############################################################################
class PECIntegrator(Integrator):
    r"""
    In the Predict-Evaluate-Correct (PEC) mode, the system is advanced using:

    .. math::

        y^{n+\frac{1}{2}} = y^n + \frac{\Delta t}{2}F(y^{n-\frac{1}{2}})
        --> Predict

        F(y^{n+\frac{1}{2}}) --> Evaluate

        y^{n + 1} = y^n + \Delta t F(y^{n+\frac{1}{2}})

    """
    def one_timestep(self, t, dt):
        self.initialize()

        # Predict
        self.stage1()
        self.update_domain()

        # Call any post-stage functions.
        self.do_post_stage(0.5*dt, 1)

        self.compute_accelerations()

        # Correct
        self.stage2()
        self.update_domain()

        # Call any post-stage functions.
        self.do_post_stage(dt, 2)


###############################################################################
# `EPECIntegrator` class
###############################################################################
class EPECIntegrator(Integrator):
    r"""
    Predictor corrector integrators can have two modes of
    operation.

    In the Evaluate-Predict-Evaluate-Correct (EPEC) mode, the
    system is advanced using:

    .. math::

        F(y^n) --> Evaluate

        y^{n+\frac{1}{2}} = y^n + F(y^n) --> Predict

        F(y^{n+\frac{1}{2}}) --> Evaluate

        y^{n+1} = y^n + \Delta t F(y^{n+\frac{1}{2}}) --> Correct

    Notes:

    The Evaluate stage of the integrator forces a function
    evaluation. Therefore, the PEC mode is much faster but relies on
    old accelertions for the Prediction stage.

    In the EPEC mode, the final corrector can be modified to:

    .. math::

        y^{n+1} = y^n + \frac{\Delta t}{2}\left( F(y^n) +
                                F(y^{n+\frac{1}{2}}) \right)

    This would require additional storage for the accelerations.

    """
    def one_timestep(self, t, dt):
        self.initialize()

        self.compute_accelerations()

        # Predict
        self.stage1()
        self.update_domain()

        # Call any post-stage functions.
        self.do_post_stage(0.5*dt, 1)

        self.compute_accelerations()

        # Correct
        self.stage2()
        self.update_domain()

        # Call any post-stage functions.
        self.do_post_stage(dt, 2)


###############################################################################
# `TVDRK3Integrator` class
###############################################################################
class TVDRK3Integrator(Integrator):
    r"""
    In the TVD-RK3 integrator, the system is advanced using:

    .. math::

        y^{n + \frac{1}{3}} = y^n + \Delta t F( y^n )

        y^{n + \frac{2}{3}} = \frac{3}{4}y^n +
        \frac{1}{4}(y^{n + \frac{1}{3}} + \Delta t F(y^{n + \frac{1}{3}}))

        y^{n + 1} = \frac{1}{3}y^n + \frac{2}{3}(y^{n + \frac{2}{3}}
        + \Delta t F(y^{n + \frac{2}{3}}))

    """
    def one_timestep(self, t, dt):
        self.initialize()

        # stage 1
        self.compute_accelerations()
        self.stage1()
        self.update_domain()
        self.do_post_stage(1./3*dt, 1)

        # stage 2
        self.compute_accelerations()
        self.stage2()
        self.update_domain()
        self.do_post_stage(2./3*dt, 2)

        # stage 3 and end
        self.compute_accelerations()
        self.stage3()
        self.update_domain()
        self.do_post_stage(dt, 3)


###############################################################################
class LeapFrogIntegrator(PECIntegrator):
    r"""A leap-frog integrator.
    """

    def one_timestep(self, t, dt):

        self.stage1()
        self.update_domain()
        self.do_post_stage(0.5*dt, 1)

        self.compute_accelerations()
        self.stage2()
        self.update_domain()
        self.do_post_stage(dt, 2)


###############################################################################
class PEFRLIntegrator(Integrator):
    r"""A Position-Extended Forest-Ruth-Like integrator [Omeylan2002]_

    References
    ----------
    .. [Omeylan2002] I.M. Omelyan, I.M. Mryglod and R. Folk, "Optimized
       Forest-Ruth- and Suzuki-like algorithms for integration of motion
       in many-body systems", Computer Physics Communications 146, 188 (2002)
       http://arxiv.org/abs/cond-mat/0110585

    """

    def one_timestep(self, t, dt):

        self.stage1()
        self.update_domain()
        self.do_post_stage(0.1786178958448091*dt, 1)

        self.compute_accelerations()
        self.stage2()
        self.update_domain()
        self.do_post_stage(0.1123533131749906*dt, 2)

        self.compute_accelerations()
        self.stage3()
        self.update_domain()
        self.do_post_stage(0.8876466868250094*dt, 3)

        self.compute_accelerations()
        self.stage4()
        self.update_domain()
        self.do_post_stage(0.8213821041551909*dt, 4)

        self.compute_accelerations()
        self.stage5()
        self.update_domain()
        self.do_post_stage(dt, 5)



"""
SPHysics case3 - wavemaker in beach (17 minutes)
"""

from pysph.base.kernels import CubicSpline
from pysph.solver.application import Application

from pysph.sph.integrator_step import WCSPHStep
from pysph.sph.integrator_step import TwoStageRigidBodyStep
from pysph.base.utils import get_particle_array

import numpy as np
from pysph.sph.scheme import AdamiHuAdamsScheme, WCSPHScheme, SchemeChooser
from pysph.sph.wc.edac import EDACScheme
from pysph.tools.geometry import remove_overlap_particles
from pysph.tools.geometry import get_2d_block
from pysph.examples.sphysics.beach_geometry import get_beach_geometry_2d


def get_wavespaddle_geometry(hdx=1.5, dx_f=0.1, dx_s=0.1, r_f=100., r_s=100.,
                             length=3.75, height=0.3, flat_l=1., angle=4.2364,
                             h_fluid=0.2):
    x1, y1, x2, y2 = get_beach_geometry_2d(dx_s, length, height, flat_l,
                                           angle, 5)
    r1 = np.ones_like(x1) * r_s
    m1 = r1 * dx_s * dx_s
    h1 = np.ones_like(x1) * hdx * dx_s
    wall = get_particle_array(name='wall', x=x1, y=y1, rho=r1, m=m1, h=h1)
    r2 = np.ones_like(x2) * r_s
    m2 = r2 * dx_s * dx_s
    h2 = np.ones_like(x2) * hdx * dx_s
    paddle = get_particle_array(name='paddle', x=x2, y=y2, rho=r2, m=m2, h=h2)
    fluid_center = np.array([flat_l - length / 2.0, h_fluid / 2.0])
    x_fluid, y_fluid = get_2d_block(dx_f, length, h_fluid, fluid_center)
    x3 = []
    y3 = []
    theta = np.pi * angle / 180.0
    for i, xi in enumerate(x_fluid):
        if y_fluid[i] >= np.tan(-theta) * xi:
            x3.append(xi)
            y3.append(y_fluid[i])
    x3 = np.array(x3)
    y3 = np.array(y3)
    r3 = np.ones_like(x3) * r_f
    m3 = r3 * dx_f * dx_f
    h3 = np.ones_like(x3) * hdx * dx_f
    fluid = get_particle_array(name='fluid', x=x3, y=y3, rho=r3, m=m3, h=h3)
    remove_overlap_particles(fluid, wall, dx_s, 2)
    remove_overlap_particles(fluid, paddle, dx_s, 2)
    return fluid, wall, paddle


class WavesPaddle2D(Application):

    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.3,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=0.01,
            help="spacing between the particles")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h0 = self.hdx * self.dx
        self.dt = 0.25 * self.h0 / co

    def pre_step(self, solver):
        t = solver.t
        theta = 2.0 * np.pi * t / period
        paddle = self.particles[2]
        paddle.u = amplitude * (paddle.y - self.dx) * np.cos(theta)
        paddle.v = amplitude * (flat_l - paddle.x) * np.cos(theta)

    def create_particles(self):
        fluid, wall, paddle = get_wavespaddle_geometry(
            self.hdx, self.dx, self.dx, h_fluid=h_fluid)
        self.scheme.setup_properties([fluid, wall, paddle])
        scheme = self.options.scheme
        if scheme == 'aha' or scheme == 'edac':
            for p in ['u0', 'v0', 'w0', 'x0', 'y0', 'z0']:
                paddle.add_property(p)
        particles = [fluid, wall, paddle]
        return particles

    def create_scheme(self):
        aha = AdamiHuAdamsScheme(['fluid'], ['wall', 'paddle'], dim=2,
                                 rho0=ro, c0=co, alpha=alp, gy=-9.81, nu=0.0,
                                 h0=0.01, gamma=1.0)
        wcsph = WCSPHScheme(['fluid'], ['wall', 'paddle'], dim=2, rho0=ro,
                            c0=co, h0=0.01, hdx=1.3, hg_correction=True,
                            gy=-9.81, alpha=alp, gamma=gamma, update_h=True)
        edac = EDACScheme(['fluid'], ['wall', 'paddle'], dim=2, rho0=ro,
                          c0=co, gy=-9.81, alpha=0.0, nu=0.0, h=0.01,
                          clamp_p=True)
        return SchemeChooser(default='wcsph', wcsph=wcsph, aha=aha, edac=edac)

    def configure_scheme(self):
        s = self.scheme
        scheme = self.options.scheme
        if scheme == 'wcsph':
            s.configure(h0=self.h0, hdx=self.hdx)
        elif scheme == 'aha':
            s.configure(h0=self.h0)
        elif scheme == 'edac':
            s.configure(h=self.h0)
        step = dict(paddle=TwoStageRigidBodyStep())
        s.configure_solver(
            kernel=CubicSpline(dim=2), tf=5.0, dt=self.dt,
            adaptive_timestep=False, extra_steppers=step)


if __name__ == '__main__':
    h_fluid = 0.25
    co = 10.0 * np.sqrt(2.0 * 9.81 * h_fluid)
    flat_l = 1.0
    gamma = 7.0
    ro = 100.0
    alp = 0.2
    amplitude = 1.0
    period = 1.4
    app = WavesPaddle2D()
    app.run()



"""Dam break past an obstacle with data from SPHysics. (40 minutes)

For benchmarking, we use the input geometry and discretization as the
SPHYSICS Case 5
(https://wiki.manchester.ac.uk/sphysics/index.php/SPHYSICS_Home_Page)

We only require the INDAT and IPART files generated by SPHysics. These
define respectively, the numerical parameters and the initial particle
data used for the run. The rest of the problem is set-up in the usual
way.

"""
import os
import numpy

from pysph.sph.equation import Group
from pysph.base.kernels import CubicSpline
from pysph.sph.wc.basic import TaitEOS, TaitEOSHGCorrection, MomentumEquation
from pysph.sph.basic_equations import ContinuityEquation, XSPHCorrection

from pysph.solver.solver import Solver
from pysph.solver.application import Application
from pysph.sph.integrator import EPECIntegrator, PECIntegrator
from pysph.sph.integrator_step import WCSPHStep

from pysph.tools.sphysics import sphysics2pysph

MY_DIR = os.path.dirname(__file__)
INDAT = os.path.join(MY_DIR, 'INDAT.gz')
IPART = os.path.join(MY_DIR, 'IPART.gz')

# problem dimensionality
dim = 3

# suggested initial time step and final time
dt = 1e-5
tf = 2.0

# physical constants for the run loaded from SPHysics INDAT
indat = numpy.loadtxt(INDAT)
H = float( indat[10] )
B = float( indat[11] )
gamma = float( indat[12] )
eps = float( indat[14] )
rho0 = float( indat[15] )
alpha = float( indat[16] )
beta = 0.0
c0 = numpy.sqrt( B*gamma/rho0 )

class DamBreak3DSPhysics(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--test", action="store_true", dest="test", default=False,
            help="For use while testing of results, uses PEC integrator."
        )

    def create_particles(self):
         return sphysics2pysph(IPART, INDAT, vtk=False)

    def create_solver(self):
        kernel = CubicSpline(dim=3)

        if self.options.test:
            integrator = PECIntegrator(fluid=WCSPHStep(),boundary=WCSPHStep())
            adaptive, n_damp = False, 0
        else:
            integrator = EPECIntegrator(fluid=WCSPHStep(),boundary=WCSPHStep())
            adaptive, n_damp = True, 0

        solver = Solver(dim=dim, kernel=kernel, integrator=integrator,
                        adaptive_timestep=adaptive, tf=tf, dt=dt,
                        n_damp=n_damp)
        return solver

    def create_equations(self):
        equations = [

            # Equation of state
            Group(equations=[
                    TaitEOS(dest='fluid', sources=None,
                            rho0=rho0, c0=c0, gamma=gamma),
                    TaitEOSHGCorrection(dest='boundary', sources=None,
                                        rho0=rho0, c0=c0, gamma=gamma),

                    ], real=False),

            # Continuity Momentum and XSPH equations
            Group(equations=[
                    ContinuityEquation(dest='fluid',
                                       sources=['fluid', 'boundary']),
                    ContinuityEquation(dest='boundary', sources=['fluid']),

                    MomentumEquation(
                        dest='fluid', sources=['fluid', 'boundary'], c0=c0,
                        alpha=alpha, beta=beta, gz=-9.81,
                        tensile_correction=True),

                       # Position step with XSPH
                    XSPHCorrection(dest='fluid', sources=['fluid'], eps=eps)
                 ])
            ]
        return equations

if __name__ == '__main__':
    app = DamBreak3DSPhysics()
    app.run()
